import re
import arango

import bel.utils as utils
from bel.Config import config

from structlog import get_logger

log = get_logger()

edgestore_db_name = "edgestore"
belns_db_name = "belns"
belapi_db_name = "belapi"

edgestore_nodes_name = "nodes"  # edgestore node collection name
edgestore_edges_name = "edges"  # edgestore edge collection name
edgestore_pipeline_name = "pipeline"  # edgestore pipeline state collection name
edgestore_pipeline_errors_name = (
    "pipeline_errors"
)  # edgestore pipeline errors collection name
edgestore_pipeline_stats_name = (
    "pipeline_stats"
)  # edgestore pipeline stats collection name

equiv_nodes_name = "equivalence_nodes"  # equivalence node collection name
equiv_edges_name = "equivalence_edges"  # equivalence edge collection name
ortholog_nodes_name = "ortholog_nodes"  # ortholog node collection name
ortholog_edges_name = "ortholog_edges"  # ortholog edge collection name
belns_metadata_name = "resources_metadata"  # BEL Resources metadata

belapi_settings_name = "settings"  # BEL API settings and configuration
belapi_statemgmt_name = "state_mgmt"  # BEL API state mgmt


# TODO - update get db and get collections using same pattern as in userstore/common/db.py
#        I made the mistake below of edgestore_db = sys_db.create_database()
#        instead of
#           sys_db.create_database('edgestore')
#           edgestore_db = client.db('edgestore')
#           if edgestore_db.has_collection('xxx'):
#               xxx_coll = edgestore_db.collection('xxx')
#           else:
#               xxx_coll = edgestore_db.create_collection('xxx')


def get_user_creds(username, password):
    """Get username/password

    Use provided username and password OR in config OR blank in that order
    """
    username = utils.first_true(
        [username, config["bel_api"]["servers"]["arangodb_username"]], default=""
    )
    password = utils.first_true(
        [password, config["secrets"]["bel_api"]["servers"].get("arangodb_password")],
        default="",
    )

    return username, password


def get_client(host=None, port=None, username=None, password=None, enable_logging=True):
    """Get arango client and edgestore db handle"""

    host = utils.first_true(
        [host, config["bel_api"]["servers"]["arangodb_host"], "localhost"]
    )
    port = utils.first_true([port, config["bel_api"]["servers"]["arangodb_port"], 8529])
    username = utils.first_true(
        [username, config["bel_api"]["servers"]["arangodb_username"], ""]
    )
    password = utils.first_true(
        [
            password,
            config.get(
                "secrets",
                config["secrets"]["bel_api"]["servers"].get("arangodb_password"),
            ),
            "",
        ]
    )

    client = arango.client.ArangoClient(
        protocol=config["bel_api"]["servers"]["arangodb_protocol"], host=host, port=port
    )

    return client


def aql_query(db, query):
    """Run AQL query"""

    result = db.aql.execute(query)
    return result


def get_edgestore_handle(
    client: arango.client.ArangoClient,
    username=None,
    password=None,
    edgestore_db_name: str = edgestore_db_name,
    edgestore_edges_name: str = edgestore_edges_name,
    edgestore_nodes_name: str = edgestore_nodes_name,
    edgestore_pipeline_name: str = edgestore_pipeline_name,
    edgestore_pipeline_stats_name: str = edgestore_pipeline_stats_name,
    edgestore_pipeline_errors_name: str = edgestore_pipeline_errors_name,
) -> arango.database.StandardDatabase:
    """Get Edgestore arangodb database handle

    Args:
        client (arango.client.ArangoClient): Description
        username (None, optional): Description
        password (None, optional): Description
        edgestore_db_name (str, optional): Description
        edgestore_edges_name (str, optional): Description
        edgestore_nodes_name (str, optional): Description

    Returns:
        arango.database.StandardDatabase: Description
    """

    (username, password) = get_user_creds(username, password)

    sys_db = client.db("_system", username=username, password=password)

    # Create a new database named "edgestore"
    try:
        if username and password:
            edgestore_db = sys_db.create_database(
                name=edgestore_db_name,
                users=[{"username": username, "password": password, "active": True}],
            )
        else:
            edgestore_db = sys_db.create_database(name=edgestore_db_name)
    except arango.exceptions.DatabaseCreateError:
        if username and password:
            edgestore_db = client.db(
                edgestore_db_name, username=username, password=password
            )
        else:
            edgestore_db = client.db(edgestore_db_name)

    # TODO - add a skiplist index for _from? or _key? to be able to do paging?
    # has_collection function doesn't seem to be working
    # if not edgestore_db.has_collection(edgestore_nodes_name):
    try:
        nodes = edgestore_db.create_collection(
            edgestore_nodes_name, index_bucket_count=64
        )
        nodes.add_hash_index(fields=["name"], unique=False)
        nodes.add_hash_index(
            fields=["components"], unique=False
        )  # add subject/object components as node properties
    except Exception:
        pass

    # if not edgestore_db.has_collection(edgestore_edges_name):
    try:
        edges = edgestore_db.create_collection(
            edgestore_edges_name, edge=True, index_bucket_count=64
        )
        edges.add_hash_index(fields=["relation"], unique=False)
        edges.add_hash_index(fields=["edge_types"], unique=False)
        edges.add_hash_index(fields=["nanopub_id"], unique=False)
        edges.add_hash_index(fields=["metadata.project"], unique=False)
        edges.add_hash_index(fields=["annotations[*].id"], unique=False)
    except Exception:
        pass

    # if not edgestore_db.has_collection(edgestore_pipeline_name):
    try:
        edgestore_db.create_collection(edgestore_pipeline_name)
    except Exception:
        pass

    try:
        edgestore_db.create_collection(edgestore_pipeline_errors_name)
    except Exception:
        pass

    try:
        edgestore_db.create_collection(edgestore_pipeline_stats_name)
    except arango.exceptions.CollectionCreateError as e:
        pass

    return edgestore_db


def get_belns_handle(client, username=None, password=None):
    """Get BEL namespace arango db handle"""

    (username, password) = get_user_creds(username, password)

    sys_db = client.db("_system", username=username, password=password)

    # Create a new database named "belns"
    try:
        if username and password:
            belns_db = sys_db.create_database(
                name=belns_db_name,
                users=[{"username": username, "password": password, "active": True}],
            )
        else:
            belns_db = sys_db.create_database(name=belns_db_name)
    except arango.exceptions.DatabaseCreateError:
        if username and password:
            belns_db = client.db(belns_db_name, username=username, password=password)
        else:
            belns_db = client.db(belns_db_name)

    try:
        belns_db.create_collection(belns_metadata_name)
    except Exception:
        pass

    try:
        equiv_nodes = belns_db.create_collection(
            equiv_nodes_name, index_bucket_count=64
        )
        equiv_nodes.add_hash_index(fields=["name"], unique=True)
    except Exception:
        pass

    try:
        belns_db.create_collection(equiv_edges_name, edge=True, index_bucket_count=64)
    except Exception:
        pass

    try:
        ortholog_nodes = belns_db.create_collection(
            ortholog_nodes_name, index_bucket_count=64
        )
        ortholog_nodes.add_hash_index(fields=["name"], unique=True)
    except Exception:
        pass

    try:
        belns_db.create_collection(
            ortholog_edges_name, edge=True, index_bucket_count=64
        )
    except Exception:
        pass

    return belns_db


def get_belapi_handle(client, username=None, password=None):
    """Get BEL API arango db handle"""

    (username, password) = get_user_creds(username, password)

    sys_db = client.db("_system", username=username, password=password)

    # Create a new database named "belapi"
    try:
        if username and password:
            belapi_db = sys_db.create_database(
                name=belapi_db_name,
                users=[{"username": username, "password": password, "active": True}],
            )
        else:
            belapi_db = sys_db.create_database(name=belapi_db_name)
    except arango.exceptions.DatabaseCreateError:
        if username and password:
            belapi_db = client.db(belapi_db_name, username=username, password=password)
        else:
            belapi_db = client.db(belapi_db_name)

    try:
        belapi_db.create_collection(belapi_settings_name)
    except Exception:
        pass

    try:
        belapi_db.create_collection(belapi_statemgmt_name)
    except Exception:
        pass

    return belapi_db


def delete_database(client, db_name, username=None, password=None):
    """Delete Arangodb database

    """

    (username, password) = get_user_creds(username, password)

    sys_db = client.db("_system", username=username, password=password)

    try:
        return sys_db.delete_database(db_name)
    except Exception:
        log.warn("No arango database {db_name} to delete, does not exist")


def batch_load_docs(db, doc_iterator, on_duplicate="replace"):
    """Batch load documents

    Args:
        db: ArangoDB client database handle
        doc_iterator:  function that yields (collection_name, doc_key, doc)
        on_duplicate: defaults to replace, but can be error, update, replace or ignore

        https://python-driver-for-arangodb.readthedocs.io/en/master/specs.html?highlight=import_bulk#arango.collection.StandardCollection.import_bulk
    """

    batch_size = 100

    counter = 0
    collections = {}
    docs = {}

    if on_duplicate not in ["error", "update", "replace", "ignore"]:
        log.error(f"Bad parameter for on_duplicate: {on_duplicate}")
        return

    for (collection_name, doc) in doc_iterator:
        if collection_name not in collections:
            collections[collection_name] = db.collection(collection_name)
            docs[collection_name] = []

        counter += 1

        docs[collection_name].append(doc)

        if counter % batch_size == 0:
            log.info(f"Bulk import arangodb: {counter}")
            for cname in docs:
                collections[cname].import_bulk(
                    docs[cname], on_duplicate=on_duplicate, halt_on_error=False
                )
                docs[cname] = []

    log.info(f"Bulk import arangodb: {counter}")
    for cname in docs:
        collections[cname].import_bulk(
            docs[cname], on_duplicate=on_duplicate, halt_on_error=False
        )
        docs[cname] = []


def arango_id_to_key(_id):
    """Remove illegal chars from potential arangodb _key (id)

    Args:
        _id (str): id to be used as arangodb _key

    Returns:
        (str): _key value with illegal chars removed
    """

    key = re.sub(r"[^a-zA-Z0-9\_\-\:\.\@\(\)\+\,\=\;\$\!\*\%]+", r"_", _id)
    if len(key) > 254:
        log.error(
            f"Arango _key cannot be longer than 254 chars: Len={len(key)}  Key: {key}"
        )
    elif len(key) < 1:
        log.error(f"Arango _key cannot be an empty string: Len={len(key)}  Key: {key}")

    return key



#!/usr/bin/env python
# -*- coding: utf-8 -*-


# Migrate BEL from 1 to 2.0.0
#

import json

import bel
from bel import BEL
import bel.lang.partialparse

from bel.Config import config
import bel.lang.ast
from bel.lang.ast import BELAst, Function, NSArg, StrArg

import structlog

log = structlog.getLogger(__name__)

bo = BEL("2.0.0", config["bel_api"]["servers"]["api_url"])
spec = bo.spec


def migrate(belstr: str) -> str:
    """Migrate BEL 1 to 2.0.0

    Args:
        bel: BEL 1

    Returns:
        bel: BEL 2
    """

    bo.ast = bel.lang.partialparse.get_ast_obj(belstr, "2.0.0")

    return migrate_ast(bo.ast).to_string()


def migrate_into_triple(belstr: str) -> str:
    """Migrate BEL1 assertion into BEL 2.0.0 SRO triple"""

    bo.ast = bel.lang.partialparse.get_ast_obj(belstr, "2.0.0")

    return migrate_ast(bo.ast).to_triple()


def migrate_ast(ast: BELAst) -> BELAst:

    # Process Subject
    bo.ast.bel_subject = convert(bo.ast.bel_subject)

    if bo.ast.bel_object:
        if bo.ast.bel_object.type == "BELAst":
            bo.ast.bel_object.bel_subject = convert(bo.ast.bel_object.bel_subject)
            if bo.ast.bel_object.bel_object:
                bo.ast.bel_object.bel_object = convert(bo.ast.bel_object.bel_object)
        else:
            bo.ast.bel_object = convert(bo.ast.bel_object)

    return bo.ast


def convert(ast):
    """Convert BEL1 AST Function to BEL2 AST Function"""

    if ast and ast.type == "Function":
        # Activity function conversion
        if (
            ast.name != "molecularActivity"
            and ast.name in spec["namespaces"]["Activity"]["list"]
        ):
            print("name", ast.name, "type", ast.type)
            ast = convert_activity(ast)
            return ast  # Otherwise - this will trigger on the BEL2 molecularActivity

        # translocation conversion
        elif ast.name in ["tloc", "translocation"]:
            ast = convert_tloc(ast)

        fus_flag = False
        for idx, arg in enumerate(ast.args):
            if arg.__class__.__name__ == "Function":

                # Fix substitution -> variation()
                if arg.name in ["sub", "substitution"]:
                    ast.args[idx] = convert_sub(arg)

                elif arg.name in ["trunc", "truncation"]:
                    ast.args[idx] = convert_trunc(arg)

                elif arg.name in ["pmod", "proteinModification"]:
                    ast.args[idx] = convert_pmod(arg)

                elif arg.name in ["fus", "fusion"]:
                    fus_flag = True

                # Recursively process Functions
                ast.args[idx] = convert(ast.args[idx])

        if fus_flag:
            ast = convert_fus(ast)

    return ast


def convert_tloc(ast):
    """Convert BEL1 tloc() to BEL2"""

    from_loc_arg = ast.args[1]
    to_loc_arg = ast.args[2]
    from_loc = Function("fromLoc", spec, parent_function=ast)
    from_loc.add_argument(
        NSArg(from_loc_arg.namespace, from_loc_arg.value, parent_function=from_loc)
    )
    to_loc = Function("toLoc", spec, parent_function=ast)
    to_loc.add_argument(
        NSArg(to_loc_arg.namespace, to_loc_arg.value, parent_function=to_loc)
    )

    ast.args[1] = from_loc
    ast.args[2] = to_loc

    return ast


def convert_activity(ast):
    """Convert BEL1 activities to BEL2 act()"""

    if len(ast.args) > 1:
        log.error(f"Activity should not have more than 1 argument {ast.to_string()}")

    p_arg = ast.args[0]  # protein argument
    print("p_arg", p_arg)
    ma_arg = Function("ma", bo.spec)
    ma_arg.add_argument(StrArg(ast.name, ma_arg))
    p_arg.change_parent_fn(ma_arg)
    ast = Function("activity", bo.spec)
    p_arg.change_parent_fn(ast)
    ast.add_argument(p_arg)
    ast.add_argument(ma_arg)

    return ast


def convert_pmod(pmod):
    """Update BEL1 pmod() protein modification term"""

    if pmod.args[0].value in spec["bel1_migration"]["protein_modifications"]:
        pmod.args[0].value = spec["bel1_migration"]["protein_modifications"][
            pmod.args[0].value
        ]

    return pmod


def convert_fus(ast):
    """Convert BEL1 fus() to BEL2 fus()"""

    parent_fn_name = ast.name_short
    prefix_list = {"p": "p.", "r": "r.", "g": "c."}
    prefix = prefix_list[parent_fn_name]

    fus1_ns = ast.args[0].namespace
    fus1_val = ast.args[0].value

    arg_fus = ast.args[1]
    fus_args = [None, "?", "?"]
    for idx, arg in enumerate(arg_fus.args):
        fus_args[idx] = arg

    fus2_ns = fus_args[0].namespace
    fus2_val = fus_args[0].value

    if fus_args[1] == "?":
        fus1_range = fus_args[1]
    else:
        fus1_range = f'"{prefix}1_{fus_args[1].value}"'

    if fus_args[2] == "?":
        fus2_range = fus_args[2]
    else:
        fus2_range = f'"{prefix}{fus_args[2].value}_?"'

    fus = Function("fus", spec, parent_function=ast)
    fus.args = [
        NSArg(fus1_ns, fus1_val, fus),
        StrArg(fus1_range, fus),
        NSArg(fus2_ns, fus2_val, fus),
        StrArg(fus2_range, fus),
    ]

    # Remove BEL
    ast_args = ast.args
    ast_args.pop(0)
    ast_args.pop(0)

    if ast_args == [None]:
        ast_args = []

    ast.args = []
    ast.add_argument(fus)

    if len(ast_args) > 0:
        ast.args.extend(ast_args)

    return ast


def convert_sub(sub):
    """Convert BEL1 sub() to BEL2 var()"""

    args = sub.args
    (ref_aa, pos, new_aa) = args

    parent_fn_name = sub.parent_function.name_short
    prefix_list = {"p": "p.", "r": "r.", "g": "c."}
    prefix = prefix_list[parent_fn_name]

    new_var_arg = f'"{prefix}{spec["namespaces"]["AminoAcid"]["to_short"][ref_aa.value]}{pos.value}{spec["namespaces"]["AminoAcid"]["to_short"][new_aa.value]}"'

    new_var = bel.lang.ast.Function("var", bo.spec)

    new_var.add_argument(StrArg(new_var_arg, new_var))

    return new_var


def convert_trunc(trunc):
    """Convert BEL1 trunc() to BEL2 var()"""

    parent_fn_name = trunc.parent_function.name_short
    prefix_list = {"p": "p.", "r": "r.", "g": "c."}
    prefix = prefix_list[parent_fn_name]

    new_var_arg = f'"truncated at {trunc.args[0].value}"'

    new_var = bel.lang.ast.Function("var", bo.spec)

    new_var.add_argument(StrArg(new_var_arg, new_var))

    return new_var


def main():

    import bel.lang.migrate_1_2

    bel1 = "kin(p(HGNC:BRAF))"

    bel1 = "p(HGNC:PIK3CA, sub(E, 545, K))"
    # bel2 = 'p(HGNC:PIK3CA, var(p.Glu545Lys))'

    bel1 = "r(HGNC:BCR, fus(HGNC:JAK2, 1875, 2626), pmod(P))"
    bel2 = 'r(fus(HGNC:BCR, "r.1_1875", HGNC:JAK2, "r.2626_?"), pmod(Ph))'

    # bel1 = 'p(HGNC:MAPK1, pmod(P, Thr, 185))'
    # bel2 = 'p(HGNC:MAPK1, pmod(Ph, Thr, 185))'

    # bel1 = 'tloc(p(HGNC:EGFR), MESHCL:Cytoplasm, MESHCL:"Cell Nucleus")'
    # bel2 = 'tloc(p(HGNC:EGFR), fromLoc(MESHCL:Cytoplasm), toLoc(MESHCL:"Cell Nucleus"))'

    # bel1 = 'p(HGNC:ABCA1, trunc(1851))'
    # bel2 = 'p(HGNC:ABCA1, var("truncated at 1851"))'

    bel2 = bel.lang.migrate_1_2.migrate(bel1)

    print("BEL2", bel2)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Pubmed related utilities

Given PMID - collect Pubmed data and Pubtator Bioconcepts used for the BELMgr
or enhancing BEL Nanopubs
"""

from typing import Mapping, Any
from lxml import etree
import re
import copy
import datetime

from bel.Config import config
import bel.lang.bel_utils as bel_utils
from bel.utils import get_url, url_path_param_quoting

import structlog

log = structlog.getLogger(__name__)

# Replace PMID
if config["bel_api"]["servers"].get("pubmed_api_key", False):
    PUBMED_TMPL = f'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&retmode=xml&api_key={config["bel_api"]["servers"]["pubmed_api_key"]}&id=PMID'
else:
    PUBMED_TMPL = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&retmode=xml&id=PMID"

PUBTATOR_TMPL = "https://www.ncbi.nlm.nih.gov/CBBresearch/Lu/Demo/RESTful/tmTool.cgi/BioConcept/PMID/JSON"

pubtator_ns_convert = {
    "CHEBI": "CHEBI",
    "Species": "TAX",
    "Gene": "EG",
    "Chemical": "MESH",
    "Disease": "MESH",
}
pubtator_entity_convert = {
    "Chemical": "Abundance",
    "Gene": "Gene",
    "Disease": "Pathology",
}
pubtator_annotation_convert = {"Disease": "Pathology"}
pubtator_known_types = [key for key in pubtator_ns_convert.keys()]


def node_text(node):
    """Needed for things like abstracts which have internal tags (see PMID:27822475)"""

    if node.text:
        result = node.text
    else:
        result = ""
    for child in node:
        if child.tail is not None:
            result += child.tail
    return result


def get_pubtator(pmid):
    """Get Pubtator Bioconcepts from Pubmed Abstract

    Re-configure the denotations into an annotation dictionary format
    and collapse duplicate terms so that their spans are in a list.
    """
    r = get_url(PUBTATOR_TMPL.replace("PMID", pmid), timeout=10)
    if r and r.status_code == 200:
        pubtator = r.json()[0]
    else:
        log.error(
            f"Cannot access Pubtator, status: {r.status_code} url: {PUBTATOR_TMPL.replace('PMID', pmid)}"
        )
        return None

    known_types = ["CHEBI", "Chemical", "Disease", "Gene", "Species"]

    for idx, anno in enumerate(pubtator["denotations"]):
        s_match = re.match(r"(\w+):(\w+)", anno["obj"])
        c_match = re.match(r"(\w+):(\w+):(\w+)", anno["obj"])
        if c_match:
            (ctype, namespace, cid) = (
                c_match.group(1),
                c_match.group(2),
                c_match.group(3),
            )

            if ctype not in known_types:
                log.info(f"{ctype} not in known_types for Pubtator")
            if namespace not in known_types:
                log.info(f"{namespace} not in known_types for Pubtator")

            pubtator["denotations"][idx][
                "obj"
            ] = f'{pubtator_ns_convert.get(namespace, "UNKNOWN")}:{cid}'
            pubtator["denotations"][idx]["entity_type"] = pubtator_entity_convert.get(
                ctype, None
            )
            pubtator["denotations"][idx][
                "annotation_type"
            ] = pubtator_annotation_convert.get(ctype, None)
        elif s_match:
            (ctype, cid) = (s_match.group(1), s_match.group(2))

            if ctype not in known_types:
                log.info(f"{ctype} not in known_types for Pubtator")

            pubtator["denotations"][idx][
                "obj"
            ] = f'{pubtator_ns_convert.get(ctype, "UNKNOWN")}:{cid}'
            pubtator["denotations"][idx]["entity_type"] = pubtator_entity_convert.get(
                ctype, None
            )
            pubtator["denotations"][idx][
                "annotation_type"
            ] = pubtator_annotation_convert.get(ctype, None)

    annotations = {}
    for anno in pubtator["denotations"]:
        log.info(anno)
        if anno["obj"] not in annotations:
            annotations[anno["obj"]] = {"spans": [anno["span"]]}
            annotations[anno["obj"]]["entity_types"] = [anno.get("entity_type", [])]
            annotations[anno["obj"]]["annotation_types"] = [
                anno.get("annotation_type", [])
            ]

        else:
            annotations[anno["obj"]]["spans"].append(anno["span"])

    del pubtator["denotations"]
    pubtator["annotations"] = copy.deepcopy(annotations)

    return pubtator


def process_pub_date(year, mon, day):
    """Create pub_date from what Pubmed provides in Journal PubDate entry
    """

    pub_date = None
    if year and re.match("[a-zA-Z]+", mon):
        pub_date = datetime.datetime.strptime(
            f"{year}-{mon}-{day}", "%Y-%b-%d"
        ).strftime("%Y-%m-%d")
    elif year:
        pub_date = f"{year}-{mon}-{day}"

    return pub_date


def get_pubmed(pmid: str) -> Mapping[str, Any]:
    """Get pubmed xml for pmid and convert to JSON

    Remove MESH terms if they are duplicated in the compound term set

    ArticleDate vs PubDate gets complicated: https://www.nlm.nih.gov/bsd/licensee/elements_descriptions.html see <ArticleDate> and <PubDate>
    Only getting pub_year at this point from the <PubDate> element.

    Args:
        pmid: pubmed id number as a string

    Returns:
        pubmed json
    """
    pubmed_url = PUBMED_TMPL.replace("PMID", str(pmid))
    r = get_url(pubmed_url)
    log.info(f"Getting Pubmed URL {pubmed_url}")

    try:
        root = etree.fromstring(r.content)
        doc = {"abstract": ""}
        doc["pmid"] = root.xpath("//PMID/text()")[0]
        doc["title"] = next(iter(root.xpath("//ArticleTitle/text()")), "")

        # TODO https://stackoverflow.com/questions/4770191/lxml-etree-element-text-doesnt-return-the-entire-text-from-an-element
        atext = next(iter(root.xpath("//Abstract/AbstractText/text()")), "")
        print("Text", atext)

        for abstracttext in root.xpath("//Abstract/AbstractText"):
            abstext = node_text(abstracttext)

            label = abstracttext.get("Label", None)
            if label:
                doc["abstract"] += f"{label}: {abstext}\n"
            else:
                doc["abstract"] += f"{abstext}\n"

        doc["abstract"] = doc["abstract"].rstrip()

        doc["authors"] = []
        for author in root.xpath("//Author"):
            last_name = next(iter(author.xpath("LastName/text()")), "")
            first_name = next(iter(author.xpath("ForeName/text()")), "")
            initials = next(iter(author.xpath("Initials/text()")), "")
            if not first_name and initials:
                first_name = initials
            doc["authors"].append(f"{last_name}, {first_name}")

        pub_year = next(
            iter(root.xpath("//Journal/JournalIssue/PubDate/Year/text()")), None
        )
        pub_mon = next(
            iter(root.xpath("//Journal/JournalIssue/PubDate/Month/text()")), "Jan"
        )
        pub_day = next(
            iter(root.xpath("//Journal/JournalIssue/PubDate/Day/text()")), "01"
        )

        pub_date = process_pub_date(pub_year, pub_mon, pub_day)

        doc["pub_date"] = pub_date
        doc["journal_title"] = next(iter(root.xpath("//Journal/Title/text()")), "")
        doc["joural_iso_title"] = next(
            iter(root.xpath("//Journal/ISOAbbreviation/text()")), ""
        )
        doc["doi"] = next(iter(root.xpath('//ArticleId[@IdType="doi"]/text()')), None)

        doc["compounds"] = []
        for chem in root.xpath("//ChemicalList/Chemical/NameOfSubstance"):
            chem_id = chem.get("UI")
            doc["compounds"].append({"id": f"MESH:{chem_id}", "name": chem.text})

        compounds = [cmpd["id"] for cmpd in doc["compounds"]]
        doc["mesh"] = []
        for mesh in root.xpath("//MeshHeading/DescriptorName"):
            mesh_id = f"MESH:{mesh.get('UI')}"
            if mesh_id in compounds:
                continue
            doc["mesh"].append({"id": mesh_id, "name": mesh.text})

        return doc
    except Exception as e:
        log.error(
            f"Bad Pubmed request, status: {r.status_code} error: {e}",
            url=f'{PUBMED_TMPL.replace("PMID", pmid)}',
        )
        return {"message": f"Cannot get PMID: {pubmed_url}"}


def enhance_pubmed_annotations(pubmed: Mapping[str, Any]) -> Mapping[str, Any]:
    """Enhance pubmed namespace IDs

    Add additional entity and annotation types to annotations
    Use preferred id for namespaces as needed
    Add strings from Title, Abstract matching Pubtator BioConcept spans

    NOTE - basically duplicated code with bel_api:api.services.pubmed

    Args:
        pubmed

    Returns:
        pubmed object
    """

    text = pubmed["title"] + pubmed["abstract"]

    annotations = {}

    for nsarg in pubmed["annotations"]:
        url = f'{config["bel_api"]["servers"]["api_url"]}/terms/{url_path_param_quoting(nsarg)}'
        log.info(f"URL: {url}")
        r = get_url(url)
        log.info(f"Result: {r}")
        new_nsarg = ""
        if r and r.status_code == 200:
            term = r.json()
            new_nsarg = bel_utils.convert_nsarg(term["id"], decanonicalize=True)

            pubmed["annotations"][nsarg]["name"] = term["name"]
            pubmed["annotations"][nsarg]["label"] = term["label"]
            pubmed["annotations"][nsarg]["entity_types"] = list(
                set(
                    pubmed["annotations"][nsarg]["entity_types"]
                    + term.get("entity_types", [])
                )
            )
            pubmed["annotations"][nsarg]["annotation_types"] = list(
                set(
                    pubmed["annotations"][nsarg]["annotation_types"]
                    + term.get("annotation_types", [])
                )
            )

        if new_nsarg != nsarg:
            annotations[new_nsarg] = copy.deepcopy(pubmed["annotations"][nsarg])
        else:
            annotations[nsarg] = copy.deepcopy(pubmed["annotations"][nsarg])

    for nsarg in annotations:
        for idx, span in enumerate(annotations[nsarg]["spans"]):
            string = text[span["begin"] - 1 : span["end"] - 1]
            annotations[nsarg]["spans"][idx]["text"] = string

    pubmed["annotations"] = copy.deepcopy(annotations)

    return pubmed


def get_pubmed_for_beleditor(pmid: str) -> Mapping[str, Any]:
    """Get fully annotated pubmed doc with Pubtator and full entity/annotation_types

    Args:
        pmid: Pubmed PMID

    Returns:
        Mapping[str, Any]: pubmed dictionary
    """

    pubmed = get_pubmed(pmid)
    pubtator = get_pubtator(pmid)
    pubmed["annotations"] = copy.deepcopy(pubtator["annotations"])

    # Add entity types and annotation types to annotations
    pubmed = enhance_pubmed_annotations(pubmed)

    return pubmed


def main():

    pmid = "19894120"

    pubmed = get_pubmed_for_beleditor(pmid)

    import json

    print("DumpVar:\n", json.dumps(pubmed, indent=4))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Usage:  program.py <customer>

"""

import copy
import itertools
import json
import datetime
import os.path
import urllib

import bel.utils as utils
import bel.db.arangodb as arangodb
import bel.nanopub.files as files
import bel.edge.edges

import structlog

log = structlog.getLogger(__name__)

client = arangodb.get_client()
edgestore_db = arangodb.get_edgestore_handle(client)

db = {"NanopubStore": client.db("NanopubStore"), "BackboneStore": client.db("BackboneStore")}

edges_coll_name = arangodb.edgestore_edges_name
nodes_coll_name = arangodb.edgestore_nodes_name


def get_edges_for_nanopub(nanopub_id):
    query = f"""
        FOR edge IN edges
            FILTER edge.nanopub_id == "{nanopub_id}"
            LIMIT 1
            RETURN edge
    """
    try:
        result = [edge for edge in edgestore_db.aql.execute(query)]
        return result[0]
    except Exception as e:
        return None


def get_nanopub(nanopub_id, db_name, collection_name):
    query = f"""
        FOR nanopub IN {collection_name}
            FILTER nanopub._key == "{nanopub_id}"
            RETURN nanopub
    """

    result = [nanopub for nanopub in db[db_name].aql.execute(query)]
    if len(result) > 0:
        return result[0]
    else:
        log.error(f"Nanopub not found for {nanopub_id} in {db_name} :: {collection_name}")
        return {}


def process_nanopub(
    nanopub_url,
    db_name: str = "NanopubStore",
    collection_name: str = "Nanopub",
    orthologize_targets: list = [],
    override: bool = False,
):

    url_comps = urllib.parse.urlparse(nanopub_url)
    nanopub_id = os.path.basename(url_comps.path)
    # domain = url_comps.netloc

    start_time = datetime.datetime.now()

    # collect nanopub
    nanopub = get_nanopub(nanopub_id, db_name, collection_name)
    if not nanopub:
        return {
            "msg": f"Could not GET nanopub id: {nanopub_id} db_name: {db_name} collection: {collection_name}",
            "edges_cnt": 0,
            "assertions_cnt": 0,
            "success": False,
            "errors": [],
        }

    nanopub["source_url"] = nanopub_url

    end_time1 = datetime.datetime.now()
    delta_ms = f"{(end_time1 - start_time).total_seconds() * 1000:.1f}"
    log.info("Timing - Get nanopub", delta_ms=delta_ms)

    # Is nanopub in edge newer than from queue? If so, skip
    if not override:
        # collect one edge for nanopub from edgestore
        edge = get_edges_for_nanopub(nanopub_id)
        if edge:
            # check if edge nanopub is newer
            # log.info("Nanopub to Edge comparison", nanopub=nanopub, edge=edge)
            if edge["metadata"].get("gd_updateTS", None):
                if nanopub["nanopub"]["metadata"]["gd_updateTS"] <= edge["metadata"]["gd_updateTS"]:
                    return {"msg": "Nanopub older than edge nanopub", "success": True, "e": ""}

    end_time2 = datetime.datetime.now()
    delta_ms = f"{(end_time2 - end_time1).total_seconds() * 1000:.1f}"
    log.info("Timing - Get edge to check nanopub", delta_ms=delta_ms)

    results = bel.edge.edges.nanopub_to_edges(nanopub, orthologize_targets=orthologize_targets)

    end_time3 = datetime.datetime.now()
    delta_ms = f"{(end_time3 - end_time2).total_seconds() * 1000:.1f}"
    log.info("Timing - Get edges for nanopub", delta_ms=delta_ms)

    if results["success"]:
        load_edges_into_db(nanopub_id, nanopub["source_url"], edges=results["edges"])

        end_time4 = datetime.datetime.now()
        delta_ms = f"{(end_time4 - end_time3).total_seconds() * 1000:.1f}"
        log.info("Timing - Load edges into edgestore", delta_ms=delta_ms)

        delta_ms = f"{(end_time4 - start_time).total_seconds() * 1000:.1f}"
        log.info("Timing - Process edge into nanopub", delta_ms=delta_ms)

        return {
            "msg": f"Loaded {len(results['edges'])} edges into edgestore",
            "edges_cnt": len(results["edges"]),
            "assertions_cnt": len(nanopub["nanopub"]["assertions"]),
            "success": True,
            "errors": results["errors"],
        }

    else:
        return {
            "msg": f'Could not process nanopub into edges - error: {results["errors"]}',
            "edges_cnt": 0,
            "assertions_cnt": len(nanopub["nanopub"]["assertions"]),
            "success": False,
            "errors": results["errors"],
        }


def load_edges_into_db(
    nanopub_id: str,
    nanopub_url: str,
    edges: list = [],
    edges_coll_name: str = edges_coll_name,
    nodes_coll_name: str = nodes_coll_name,
):
    """Load edges into Edgestore"""

    start_time = datetime.datetime.now()

    # Clean out edges for nanopub in edgestore
    query = f"""
        FOR edge IN {edges_coll_name}
            FILTER edge.nanopub_id == "{nanopub_id}"
            REMOVE edge IN edges
    """

    try:
        edgestore_db.aql.execute(query)
    except Exception as e:
        log.debug(f"Could not remove nanopub-related edges: {query}  msg: {e}")

    end_time1 = datetime.datetime.now()
    delta_ms = f"{(end_time1 - start_time).total_seconds() * 1000:.1f}"
    log.info("Timing - Delete edges for nanopub", delta_ms=delta_ms)

    # Clean out errors for nanopub in pipeline_errors
    query = f"""
        FOR e IN pipeline_errors
            FILTER e.nanopub_id == "{nanopub_id}"
            REMOVE e IN pipeline_errors
    """
    try:
        edgestore_db.aql.execute(query)
    except Exception as e:
        log.debug(f"Could not remove nanopub-related errors: {query} msg: {e}")

    end_time2 = datetime.datetime.now()
    delta_ms = f"{(end_time2 - end_time1).total_seconds() * 1000:.1f}"
    log.info("Timing - Delete pipeline errors for nanopub", delta_ms=delta_ms)

    # Collect edges and nodes to load into arangodb
    node_list, edge_list = [], []
    for doc in edge_iterator(edges=edges):
        if doc[0] == "nodes":
            node_list.append(doc[1])
        else:
            edge_list.append(doc[1])

    end_time3 = datetime.datetime.now()
    delta_ms = f"{(end_time3 - end_time2).total_seconds() * 1000:.1f}"
    log.info("Timing - Collect edges and nodes", delta_ms=delta_ms)

    try:
        results = edgestore_db.collection(edges_coll_name).import_bulk(
            edge_list, on_duplicate="replace", halt_on_error=False
        )
    except Exception as e:
        log.error(f"Could not load edges  msg: {e}")

    end_time4 = datetime.datetime.now()
    delta_ms = f"{(end_time4 - end_time3).total_seconds() * 1000:.1f}"
    log.info("Timing - Load edges into edgestore", delta_ms=delta_ms)

    try:
        results = edgestore_db.collection(nodes_coll_name).import_bulk(
            node_list, on_duplicate="replace", halt_on_error=False
        )
    except Exception as e:
        log.error(f"Could not load nodes  msg: {e}")

    end_time5 = datetime.datetime.now()
    delta_ms = f"{(end_time5 - end_time4).total_seconds() * 1000:.1f}"
    log.info("Timing - Load nodes into edgestore", delta_ms=delta_ms)


def edge_iterator(edges=[], edges_fn=None):
    """Yield documents from edge for loading into ArangoDB"""

    for edge in itertools.chain(edges, files.read_edges(edges_fn)):

        subj = copy.deepcopy(edge["edge"]["subject"])
        subj_id = str(utils._create_hash_from_doc(subj))
        subj["_key"] = subj_id
        obj = copy.deepcopy(edge["edge"]["object"])

        obj_id = str(utils._create_hash_from_doc(obj))
        obj["_key"] = obj_id
        relation = copy.deepcopy(edge["edge"]["relation"])

        relation["_from"] = f"nodes/{subj_id}"
        relation["_to"] = f"nodes/{obj_id}"

        # Create edge _key
        relation_hash = copy.deepcopy(relation)
        relation_hash.pop("edge_dt", None)
        relation_hash.pop("edge_hash", None)
        relation_hash.pop("nanopub_dt", None)
        relation_hash.pop("nanopub_url", None)
        relation_hash.pop("subject_canon", None)
        relation_hash.pop("object_canon", None)
        relation_hash.pop("public_flag", None)
        relation_hash.pop("metadata", None)

        relation_id = str(utils._create_hash_from_doc(relation_hash))
        relation["_key"] = relation_id

        if edge.get("nanopub_id", None):
            if "metadata" not in relation:
                relation["metadata"] = {}
            relation["metadata"]["nanopub_id"] = edge["nanopub_id"]

        yield ("nodes", subj)
        yield ("nodes", obj)
        yield ("edges", relation)


def main():
    pass


if __name__ == "__main__":
    main()



import requests
import json
import urllib.parse

from arango import ArangoError

import bel.utils
from bel.Config import config
import bel.db.arangodb as arangodb

from structlog import get_logger

log = get_logger()

arango_client = arangodb.get_client()

belapi_db = arangodb.get_belapi_handle(arango_client)

state_mgmt = belapi_db.collection(arangodb.belapi_statemgmt_name)
start_dates_doc_key = "nanopubstore_start_dates"


def update_nanopubstore_start_dt(url: str, start_dt: str):
    """Add nanopubstore start_dt to belapi.state_mgmt collection

    Args:
        url: url of nanopubstore
        start_dt: datetime of last query against nanopubstore for new ID's
    """

    hostname = urllib.parse.urlsplit(url)[1]

    start_dates_doc = state_mgmt.get(start_dates_doc_key)
    if not start_dates_doc:
        start_dates_doc = {
            "_key": start_dates_doc_key,
            "start_dates": [{"nanopubstore": hostname, "start_dt": start_dt}],
        }
        state_mgmt.insert(start_dates_doc)
    else:
        for idx, start_date in enumerate(start_dates_doc["start_dates"]):
            if start_date["nanopubstore"] == hostname:
                start_dates_doc["start_dates"][idx]["start_dt"] = start_dt
                break
        else:
            start_dates_doc["start_dates"].append(
                {"nanopubstore": hostname, "start_dt": start_dt}
            )

        state_mgmt.replace(start_dates_doc)


def get_nanopubstore_start_dt(url: str):
    """Get last start_dt recorded for getting new nanopub ID's"""

    hostname = urllib.parse.urlsplit(url)[1]

    start_dates_doc = state_mgmt.get(start_dates_doc_key)
    if start_dates_doc and start_dates_doc.get("start_dates"):
        date = [
            dt["start_dt"]
            for dt in start_dates_doc["start_dates"]
            if dt["nanopubstore"] == hostname
        ]
        log.info(f"Selected start_dt: {date}  len: {len(date)}")
        if len(date) == 1:
            return date[0]

    return "1900-01-01T00:00:00.000Z"


def get_nanopub_urls(ns_root_url: str = None, start_dt: str = None) -> dict:
    """Get modified and deleted nanopub urls

    Limited by last datetime retrieved (start_dt).  Modified includes new and updated nanopubs

    Returns:
        dict: {'modified': [], 'deleted': []}
    """
    if not ns_root_url:
        ns_root_url = config["bel_api"]["servers"]["nanopubstore"]

    url = f"{ns_root_url}/nanopubs/timed"

    if not start_dt:
        start_dt = get_nanopubstore_start_dt(ns_root_url)

    params = {"startTime": start_dt, "published": True}

    # TODO - this is coming back without a status code in some cases - why?
    r = bel.utils.get_url(url, params=params, cache=False)
    if r and r.status_code == 200:
        data = r.json()
        new_start_dt = data["queryTime"]
        update_nanopubstore_start_dt(ns_root_url, new_start_dt)
        nanopub_urls = {"modified": [], "deleted": []}

        # Deleted nanopubs
        for nid in data["deleteddata"]:
            nanopub_urls["deleted"].append(f"{ns_root_url}/nanopubs/{nid}")

        # Modified nanopubs
        for nid in data["data"]:
            nanopub_urls["modified"].append(f"{ns_root_url}/nanopubs/{nid}")

        return nanopub_urls

    else:
        log.error(
            f"Bad request to Nanopubstore",
            url=url,
            status=r.status_code,
            type="api_request",
        )
        return {}


def get_nanopub(url):
    """Get Nanopub from nanopubstore given url"""

    r = bel.utils.get_url(url, cache=False)
    if r and r.json():
        return r.json()
    else:
        return {}



import click
import json
import yaml
import gzip
import re
import sys
import timy

import bel.db.arangodb
import bel.db.elasticsearch
import bel.edge.edges
import bel.utils as utils
import bel.Config
from bel.Config import config

from bel.lang.belobj import BEL

import bel.nanopub.nanopubs as bnn
import bel.nanopub.files as bnf
import bel.nanopub.belscripts

import logging
import logging.config

if config.get("logging", False):
    logging.config.dictConfig(config.get("logging"))
log = logging.getLogger(__name__)


# Add -h to help options for commands
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


class Context(object):
    def __init__(self):
        self.config = config


pass_context = click.make_pass_decorator(Context, ensure=True)


@click.group(context_settings=CONTEXT_SETTINGS)
def belc():
    """ BEL commands

    Uses first file found to load in default configuration:

        ./belbio_conf.yaml
        ./.belbio_conf
        ~/.belbio_conf
    """
    pass


@belc.group()
def nanopub():
    """Nanopub specific commands"""

    pass


@belc.command(context_settings=CONTEXT_SETTINGS)
@click.argument("input_fn")
@click.option("--db_save", help="Save Edges direct to EdgeStore")
@click.option("--db_delete", help="Delete EdgeStore prior to saving new Edges")
@click.option(
    "--output_fn", default="-", help="BEL Edges output filename - defaults to STDOUT"
)
@click.option(
    "--rules",
    help='Select specific rules to create BEL Edges, comma-delimited, e.g. "component_of,degradation", default is to run all rules. Special rule: "skip" does not compute edges at all - just processes primary edge',
)
@click.option(
    "--species", help="Orthologize to species (Format TAX:<NCBI tax_id_number>)"
)
@click.option(
    "--namespace_targets",
    help='Target namespaces for canonicalizing BEL, e.g. {"HGNC": ["EG", "SP"], "CHEMBL": ["CHEBI"]}',
)
@click.option("--version", help="BEL language version")
@click.option("--api", help="API Endpoint to use for BEL Entity validation")
@click.option(
    "--config_fn", help="BEL configuration file - overrides default configuration files"
)
@pass_context
def pipeline(
    ctx,
    input_fn,
    db_save,
    db_delete,
    output_fn,
    rules,
    species,
    namespace_targets,
    version,
    api,
    config_fn,
):
    """BEL Pipeline - BEL Nanopubs into BEL Edges

    This will process BEL Nanopubs into BEL Edges by validating, orthologizing (if requested),
    canonicalizing, and then computing the BEL Edges based on the given rule_set.

    \b
    input_fn:
        If input fn has *.gz, will read as a gzip file
        If input fn has *.jsonl*, will parsed as a JSONLines file
        IF input fn has *.json*, will be parsed as a JSON file
        If input fn has *.yaml* or *.yml*,  will be parsed as a YAML file

    \b
    output_fn:
        If output fn has *.gz, will written as a gzip file
        If output fn has *.jsonl*, will written as a JSONLines file
        IF output fn has *.json*, will be written as a JSON file
        If output fn has *.yaml* or *.yml*,  will be written as a YAML file
        If output fn has *.jgf, will be written as JSON Graph Formatted file
    """

    if config_fn:
        config = bel.db.Config.merge_config(ctx.config, override_config_fn=config_fn)
    else:
        config = ctx.config

    # Configuration - will return the first truthy result in list else the default option
    if namespace_targets:
        namespace_targets = json.loads(namespace_targets)
    if rules:
        rules = rules.replace(" ", "").split(",")

    namespace_targets = utils.first_true(
        [namespace_targets, config["bel"]["lang"].get("canonical")], None
    )
    rules = utils.first_true(
        [rules, config["bel"]["nanopub"].get("pipeline_edge_rules", False)], False
    )
    api = utils.first_true(
        [api, config["bel_api"]["servers"].get("api_url", None)], None
    )
    version = utils.first_true(
        [version, config["bel"]["lang"].get("default_bel_version", None)], None
    )

    n = bnn.Nanopub()

    try:
        json_flag, jsonl_flag, yaml_flag, jgf_flag = False, False, False, False
        all_bel_edges = []
        fout = None

        if db_save or db_delete:
            if db_delete:
                arango_client = bel.db.arangodb.get_client()
                bel.db.arangodb.delete_database(arango_client, "edgestore")
            else:
                arango_client = bel.db.arangodb.get_client()

            edgestore_handle = bel.db.arangodb.get_edgestore_handle(arango_client)

        elif re.search("ya?ml", output_fn):
            yaml_flag = True
        elif "jsonl" in output_fn:
            jsonl_flag = True
        elif "json" in output_fn:
            json_flag = True
        elif "jgf" in output_fn:
            jgf_flag = True

        if db_save:
            pass
        elif "gz" in output_fn:
            fout = gzip.open(output_fn, "wt")
        else:
            fout = open(output_fn, "wt")

        nanopub_cnt = 0
        with timy.Timer() as timer:
            for np in bnf.read_nanopubs(input_fn):
                # print('Nanopub:\n', json.dumps(np, indent=4))

                nanopub_cnt += 1
                if nanopub_cnt % 100 == 0:
                    timer.track(f"{nanopub_cnt} Nanopubs processed into Edges")

                bel_edges = n.bel_edges(
                    np,
                    namespace_targets=namespace_targets,
                    orthologize_target=species,
                    rules=rules,
                )

                if db_save:
                    bel.edge.edges.load_edges_into_db(edgestore_handle, edges=bel_edges)
                elif jsonl_flag:
                    fout.write("{}\n".format(json.dumps(bel_edges)))
                else:
                    all_bel_edges.extend(bel_edges)

        if db_save:
            pass
        elif yaml_flag:
            fout.write("{}\n".format(yaml.dumps(all_bel_edges)))
        elif json_flag:
            fout.write("{}\n".format(json.dumps(all_bel_edges)))
        elif jgf_flag:
            bnf.edges_to_jgf(output_fn, all_bel_edges)

    finally:
        if fout:
            fout.close()


@nanopub.command(name="validate", context_settings=CONTEXT_SETTINGS)
@click.option(
    "--output_fn",
    type=click.File("wt"),
    default="-",
    help="BEL Edges JSON output filename - defaults to STDOUT",
)
@click.option("--api", help="BEL.bio API endpoint")
@click.option(
    "--config_fn",
    help="BEL Pipeline configuration file - overrides default configuration files",
)
@click.argument("input_fn")
@pass_context
def nanopub_validate(ctx, input_fn, output_fn, api, config_fn):
    """Validate nanopubs"""

    if config_fn:
        config = bel.db.Config.merge_config(ctx.config, override_config_fn=config_fn)
    else:
        config = ctx.config

    api = utils.first_true(
        [api, config["bel_api"]["servers"].get("api_url", None)], None
    )

    print(f"Running validate nanopubs using {api}")

    # if target:
    #     with open(target, 'w') as h:
    #         h.write(content)
    # else:
    #     sys.stdout.write(content)


@nanopub.command(name="belscript", context_settings=CONTEXT_SETTINGS)
@click.option("--input_fn", "-i", default="-")
@click.option("--output_fn", "-o", default="-")
@pass_context
def convert_belscript(ctx, input_fn, output_fn):
    """Convert belscript to nanopubs_bel format

    This will convert the OpenBEL BELScript file format to
    nanopub_bel-1.0.0 format.

    \b
    input_fn:
        If input fn has *.gz, will read as a gzip file

    \b
    output_fn:
        If output fn has *.gz, will written as a gzip file
        If output fn has *.jsonl*, will written as a JSONLines file
        IF output fn has *.json*, will be written as a JSON file
        If output fn has *.yaml* or *.yml*,  will be written as a YAML file
    """

    try:

        (
            out_fh,
            yaml_flag,
            jsonl_flag,
            json_flag,
        ) = bel.nanopub.files.create_nanopubs_fh(output_fn)
        if yaml_flag or json_flag:
            docs = []

        # input file
        if re.search("gz$", input_fn):
            f = gzip.open(input_fn, "rt")
        else:
            f = open(input_fn, "rt")

        # process belscript
        for doc in bel.nanopub.belscripts.parse_belscript(f):
            if yaml_flag or json_flag:
                docs.append(doc)
            elif jsonl_flag:
                out_fh.write("{}\n".format(json.dumps(doc)))

        if yaml_flag:
            yaml.dump(docs, out_fh)

        elif json_flag:
            json.dump(docs, out_fh, indent=4)

    finally:
        f.close()
        out_fh.close()


@nanopub.command(name="reformat", context_settings=CONTEXT_SETTINGS)
@click.option("--input_fn", "-i")
@click.option("--output_fn", "-o")
@pass_context
def reformat(ctx, input_fn, output_fn):
    """Reformat between JSON, YAML, JSONLines formats

    \b
    input_fn:
        If input fn has *.gz, will read as a gzip file

    \b
    output_fn:
        If output fn has *.gz, will written as a gzip file
        If output fn has *.jsonl*, will written as a JSONLines file
        IF output fn has *.json*, will be written as a JSON file
        If output fn has *.yaml* or *.yml*,  will be written as a YAML file
    """

    try:

        (
            out_fh,
            yaml_flag,
            jsonl_flag,
            json_flag,
        ) = bel.nanopub.files.create_nanopubs_fh(output_fn)
        if yaml_flag or json_flag:
            docs = []

        # input file
        if re.search("gz$", input_fn):
            f = gzip.open(input_fn, "rt")
        else:
            f = open(input_fn, "rt")

        for np in bnf.read_nanopubs(input_fn):
            if yaml_flag or json_flag:
                docs.append(np)
            elif jsonl_flag:
                out_fh.write("{}\n".format(json.dumps(np)))

        if yaml_flag:
            yaml.dump(docs, out_fh)

        elif json_flag:
            json.dump(docs, out_fh, indent=4)

    finally:
        f.close()
        out_fh.close()


@nanopub.command(name="stats", context_settings=CONTEXT_SETTINGS)
@click.argument("input_fn")
@pass_context
def nanopub_stats(ctx, input_fn):
    """Collect statistics on nanopub file

    input_fn can be json, jsonl or yaml and additionally gzipped
    """

    counts = {
        "nanopubs": 0,
        "assertions": {"total": 0, "subject_only": 0, "nested": 0, "relations": {}},
    }

    for np in bnf.read_nanopubs(input_fn):
        if "nanopub" in np:
            counts["nanopubs"] += 1
            counts["assertions"]["total"] += len(np["nanopub"]["assertions"])
            for assertion in np["nanopub"]["assertions"]:
                if assertion["relation"] is None:
                    counts["assertions"]["subject_only"] += 1
                else:
                    if re.match("\s*\(", assertion["object"]):
                        counts["assertions"]["nested"] += 1

                    if (
                        not assertion.get("relation")
                        in counts["assertions"]["relations"]
                    ):
                        counts["assertions"]["relations"][assertion.get("relation")] = 1
                    else:
                        counts["assertions"]["relations"][
                            assertion.get("relation")
                        ] += 1

    counts["assertions"]["relations"] = sorted(counts["assertions"]["relations"])

    print("DumpVar:\n", json.dumps(counts, indent=4))


@belc.group()
def stmt():
    """BEL Statement specific commands"""

    pass


@stmt.command(name="validate", context_settings=CONTEXT_SETTINGS)
@click.option("--version", help="BEL language version")
@click.option("--api", help="API Endpoint to use for BEL Entity validation")
@click.option(
    "--config_fn",
    help="BEL Pipeline configuration file - overrides default configuration files",
)
@click.argument("statement")
@pass_context
def stmt_validate(ctx, statement, version, api, config_fn):
    """Parse statement and validate """

    if config_fn:
        config = bel.db.Config.merge_config(ctx.config, override_config_fn=config_fn)
    else:
        config = ctx.config

    # Configuration - will return the first truthy result in list else the default option
    api = utils.first_true(
        [api, config["bel_api"]["servers"].get("api_url", None)], None
    )
    version = utils.first_true(
        [version, config["bel"]["lang"].get("default_bel_version", None)], None
    )

    print("------------------------------")
    print("BEL version: {}".format(version))
    print("API Endpoint: {}".format(api))
    print("------------------------------")

    bo = BEL(version=version, endpoint=api)
    bo.parse(statement)

    if bo.ast is None:
        print(bo.original_bel_stmt)
        print(bo.parse_visualize_error)
        print(bo.validation_messages)
    else:
        print(bo.ast.to_triple())
        if bo.validation_messages:
            print(bo.validation_messages)
        else:
            print("No problems found")
    return


@stmt.command()
@click.option(
    "--namespace_targets",
    help='Target namespaces for canonicalizing BEL, e.g. {"HGNC": ["EG", "SP"], "CHEMBL": ["CHEBI"]}',
)
@click.option("--version", help="BEL language version")
@click.option("--api", help="API Endpoint to use for BEL Entity validation")
@click.option(
    "--config_fn",
    help="BEL Pipeline configuration file - overrides default configuration files",
)
@click.argument("statement")
@pass_context
def canonicalize(ctx, statement, namespace_targets, version, api, config_fn):
    """Canonicalize statement

    Target namespaces can be provided in the following manner:

        bel stmt canonicalize "<BELStmt>" --namespace_targets '{"HGNC": ["EG", "SP"], "CHEMBL": ["CHEBI"]}'
            the value of target_namespaces must be JSON and embedded in single quotes
            reserving double quotes for the dictionary elements
    """

    if config_fn:
        config = bel.db.Config.merge_config(ctx.config, override_config_fn=config_fn)
    else:
        config = ctx.config

    # Configuration - will return the first truthy result in list else the default option
    if namespace_targets:
        namespace_targets = json.loads(namespace_targets)

    namespace_targets = utils.first_true(
        [namespace_targets, config.get("canonical")], None
    )
    api = utils.first_true([api, config.get("api", None)], None)
    version = utils.first_true([version, config.get("bel_version", None)], None)

    print("------------------------------")
    print("BEL version: {}".format(version))
    print("API Endpoint: {}".format(api))
    print("------------------------------")

    bo = BEL(version=version, endpoint=api)
    bo.parse(statement).canonicalize(namespace_targets=namespace_targets)

    if bo.ast is None:
        print(bo.original_bel_stmt)
        print(bo.parse_visualize_error)
        print(bo.validation_messages)
    else:
        print("ORIGINAL ", bo.original_bel_stmt)
        print("CANONICAL", bo.ast)
        if bo.validation_messages:
            print(bo.validation_messages)
        else:
            print("No problems found")
    return


@stmt.command()
@click.option("--species", help="species ID format TAX:<tax_id_number>")
@click.option("--version", help="BEL language version")
@click.option("--api", help="API Endpoint to use for BEL Entity validation")
@click.option(
    "--config_fn",
    help="BEL Pipeline configuration file - overrides default configuration files",
)
@click.argument("statement")
@pass_context
def orthologize(ctx, statement, species, version, api, config_fn):
    """Orthologize statement

    species ID needs to be the NCBI Taxonomy ID in this format: TAX:<tax_id_number>
    You can use the following common names for species id: human, mouse, rat
      (basically whatever is supported at the api orthologs endpoint)
    """

    if config_fn:
        config = bel.db.Config.merge_config(ctx.config, override_config_fn=config_fn)
    else:
        config = ctx.config

    # Configuration - will return the first truthy result in list else the default option
    api_url = utils.first_true(
        [api, config["bel_api"]["servers"].get("api_url", None)], None
    )
    version = utils.first_true(
        [version, config["bel"]["lang"].get("default_bel_version", None)], None
    )

    print("------------------------------")
    print("BEL version: {}".format(version))
    print("API Endpoint: {}".format(api))
    print("------------------------------")

    bo = BEL(version=version, endpoint=api_url)
    bo.parse(statement).orthologize(species)

    if bo.ast is None:
        print(bo.original_bel_stmt)
        print(bo.parse_visualize_error)
        print(bo.validation_messages)
    else:
        print("ORIGINAL     ", bo.original_bel_stmt)
        print("ORTHOLOGIZED ", bo.ast)
        if bo.validation_messages:
            print(bo.validation_messages)
        else:
            print("No problems found")
    return


@stmt.command()
@click.option(
    "--rules",
    help='Select specific rules to create BEL Edges, comma-delimited, e.g. "component_of,degradation", default is to run all rules',
)
@click.option("--species", help="Species ID format TAX:<tax_id_number>")
@click.option(
    "--namespace_targets",
    help='Target namespaces for canonicalizing BEL, e.g. {"HGNC": ["EG", "SP"], "CHEMBL": ["CHEBI"]}',
)
@click.option("--version", help="BEL language version")
@click.option("--api", help="API Endpoint to use for BEL Entity validation")
@click.option(
    "--config_fn",
    help="BEL Pipeline configuration file - overrides default configuration files",
)
@click.argument("statement")
@pass_context
def edges(ctx, statement, rules, species, namespace_targets, version, api, config_fn):
    """Create BEL Edges from BEL Statement"""

    if config_fn:
        config = bel.db.Config.merge_config(ctx.config, override_config_fn=config_fn)
    else:
        config = ctx.config

    # Configuration - will return the first truthy result in list else the default option
    if namespace_targets:
        namespace_targets = json.loads(namespace_targets)
    if rules:
        rules = rules.replace(" ", "").split(",")

    namespace_targets = utils.first_true(
        [namespace_targets, config["bel"]["lang"].get("canonical")], None
    )
    api_url = utils.first_true(
        [api, config["bel_api"]["servers"].get("api_url", None)], None
    )
    version = utils.first_true(
        [version, config["bel"]["lang"].get("default_bel_version", None)], None
    )

    print("------------------------------")
    print("BEL version: {}".format(version))
    print("API Endpoint: {}".format(api))
    print("------------------------------")

    bo = BEL(version=version, endpoint=api_url)
    if species:
        edges = (
            bo.parse(statement)
            .orthologize(species)
            .canonicalize(namespace_targets=namespace_targets)
            .compute_edges(rules=rules)
        )
    else:
        edges = (
            bo.parse(statement)
            .canonicalize(namespace_targets=namespace_targets)
            .compute_edges(rules=rules)
        )

    if edges is None:
        print(bo.original_bel_stmt)
        print(bo.parse_visualize_error)
        print(bo.validation_messages)
    else:
        print(json.dumps(edges, indent=4))

        if bo.validation_messages:
            print(bo.validation_messages)
        else:
            print("No problems found")
    return


@belc.group()
def db():
    """Database specific commands"""
    pass


@db.command()
@click.option(
    "--delete/--no-delete", default=False, help="Remove indexes and re-create them"
)
@click.option(
    "--index_name",
    default="terms_blue",
    help='Use this name for index. Default is "terms_blue"',
)
def elasticsearch(delete, index_name):
    """Setup Elasticsearch namespace indexes

    This will by default only create the indexes and run the namespace index mapping
    if the indexes don't exist.  The --delete option will force removal of the
    index if it exists.

    The index_name should be aliased to the index 'terms' when it's ready"""

    if delete:
        bel.db.elasticsearch.get_client(delete=True)
    else:
        bel.db.elasticsearch.get_client()


@db.command()
@click.argument("db_name")
@click.option(
    "--delete/--no-delete", default=False, help="Remove indexes and re-create them"
)
def arangodb(delete, db_name):
    """Setup ArangoDB database

    db_name: Either 'belns' or 'edgestore' - must be one or the other

    This will create the database, collections and indexes on the collection if it doesn't exist.

    The --delete option will force removal of the database if it exists."""

    if delete:
        client = bel.db.arangodb.get_client()
        bel.db.arangodb.delete_database(client, db_name)

    if db_name == "belns":
        bel.db.arangodb.get_belns_handle(client)
    elif db_name == "edgestore":
        bel.db.arangodb.get_edgestore_handle(client)



import logging.config
from structlog import configure, processors, stdlib, threadlocal

# Found in https://blog.sneawo.com/blog/2017/07/28/json-logging-in-python/

logging.config.dictConfig(
    {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "json": {
                "format": "%(message)s %(lineno)d %(pathname)s",
                "class": "pythonjsonlogger.jsonlogger.JsonFormatter",
            }
        },
        "handlers": {"json": {"class": "logging.StreamHandler", "formatter": "json"}},
        "loggers": {
            "": {"handlers": ["json"], "level": "INFO"},
            "bel": {"level": "INFO"},
            "belapi": {"level": "INFO"},
            "requests": {"level": "DEBUG"},
            "elasticsearch": {"level": "WARNING"},
            "falcon_cors": {"level": "WARNING", "formatter": "json"},
            "urllib3": {"level": "CRITICAL"},
            "timy": {"level": "ERROR"},
        },
    }
)

configure(
    context_class=threadlocal.wrap_dict(dict),
    logger_factory=stdlib.LoggerFactory(),
    wrapper_class=stdlib.BoundLogger,
    processors=[
        stdlib.filter_by_level,
        stdlib.add_logger_name,
        stdlib.add_log_level,
        stdlib.PositionalArgumentsFormatter(),
        processors.TimeStamper(fmt="iso"),
        processors.StackInfoRenderer(),
        processors.format_exc_info,
        processors.UnicodeDecoder(),
        stdlib.render_to_log_kwargs,
    ],
)


# import logging
# import logging.config
# import structlog


# def setup_logging():
#     """Setup logging for bel and bel_api"""

#     timestamper = structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S")
#     pre_chain = [
#         # Add the log level and a timestamp to the event_dict if the log entry
#         # is not from structlog.
#         structlog.stdlib.add_log_level,
#         timestamper,
#     ]

#     # logging.config.dictConfig(config['logging'])

#     logging.config.dictConfig({
#         "version": 1,
#         "disable_existing_loggers": False,
#         "formatters": {
#             "plain": {
#                 "()": structlog.stdlib.ProcessorFormatter,
#                 "processor": structlog.processors.JSONRenderer(),
#                 "foreign_pre_chain": pre_chain,
#             },
#             # "colored": {
#             #     "()": structlog.stdlib.ProcessorFormatter,
#             #     "processor": structlog.dev.ConsoleRenderer(colors=True),
#             #     "foreign_pre_chain": pre_chain,
#             # },
#         },
#         "handlers": {
#             "default": {
#                 "level": "INFO",
#                 "class": "logging.StreamHandler",
#                 "formatter": "plain",
#             },
#             "colored": {
#                 "level": "DEBUG",
#                 "class": "logging.StreamHandler",
#                 "formatter": "plain",
#             }
#         },
#         "loggers": {
#             "": {
#                 "handlers": ["default"],
#                 "level": "INFO",
#                 "propagate": True,
#             },
#             "elasticsearch": {
#                 "level": "WARNING"
#             },
#             "falcon_cors": {
#                 "level": "DEBUG",
#                 "handlers": ["default"],
#             },
#             "urllib3": {
#                 "level": "CRITICAL"
#             },
#             "timy": {
#                 "level": "ERROR"
#             },
#         }
#     })

#     structlog.configure_once(
#         processors=[
#             structlog.stdlib.filter_by_level,
#             structlog.stdlib.add_logger_name,
#             structlog.stdlib.add_log_level,
#             structlog.stdlib.PositionalArgumentsFormatter(),
#             timestamper,
#             structlog.processors.UnicodeDecoder(),
#             structlog.processors.StackInfoRenderer(),
#             structlog.processors.format_exc_info,
#             structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
#         ],
#         context_class=dict,
#         logger_factory=structlog.stdlib.LoggerFactory(),
#         wrapper_class=structlog.stdlib.BoundLogger,
#         cache_logger_on_first_use=True,
#     )


# import logging.config

# from structlog import configure, processors, stdlib, threadlocal

# logging.config.dictConfig({
#     'version': 1,
#     'disable_existing_loggers': False,
#     'formatters': {
#         'json': {
#             'format': '%(message)s %(lineno)d %(pathname)s',
#             'class': 'pythonjsonlogger.jsonlogger.JsonFormatter'
#         }
#     },
#     'handlers': {
#         'json': {
#             'class': 'logging.StreamHandler',
#             'formatter': 'json'
#         }
#     },
#     'loggers': {
#         '': {
#             'handlers': ['json'],
#             'level': logging.INFO
#         }
#     }
# })

# configure(
#     context_class=threadlocal.wrap_dict(dict),
#     logger_factory=stdlib.LoggerFactory(),
#     wrapper_class=stdlib.BoundLogger,
#     processors=[
#         stdlib.filter_by_level,
#         stdlib.add_logger_name,
#         stdlib.add_log_level,
#         stdlib.PositionalArgumentsFormatter(),
#         processors.TimeStamper(fmt="iso"),
#         processors.StackInfoRenderer(),
#         processors.format_exc_info,
#         processors.UnicodeDecoder(),
#         stdlib.render_to_log_kwargs]
# )



# BEL object utilities

import re
import json
import yaml
import copy
from typing import Mapping, List

from bel.lang.ast import NSArg
from bel.Config import config
from bel.utils import get_url, url_path_param_quoting
import bel.terms.terms
import bel.terms.orthologs

import logging

log = logging.getLogger(__name__)


def convert_nsarg_db(nsarg: str) -> dict:
    """Get default canonical and decanonical versions of nsarg

    Returns:
        dict: {'canonical': <nsarg>, 'decanonical': <nsarg>}
    """


def convert_nsarg(
    nsarg: str,
    api_url: str = None,
    namespace_targets: Mapping[str, List[str]] = None,
    canonicalize: bool = False,
    decanonicalize: bool = False,
) -> str:
    """[De]Canonicalize NSArg

    Args:
        nsarg (str): bel statement string or partial string (e.g. subject or object)
        api_url (str): BEL.bio api url to use, e.g. https://api.bel.bio/v1
        namespace_targets (Mapping[str, List[str]]): formatted as in configuration file example
        canonicalize (bool): use canonicalize endpoint/namespace targets
        decanonicalize (bool): use decanonicalize endpoint/namespace targets

    Results:
        str: converted NSArg
    """

    if not api_url:
        api_url = config["bel_api"]["servers"]["api_url"]
        if not api_url:
            log.error("Missing api url - cannot convert namespace")
            return None

    params = None
    if namespace_targets:
        namespace_targets_str = json.dumps(namespace_targets)
        params = {"namespace_targets": namespace_targets_str}

    if not namespace_targets:
        if canonicalize:
            api_url = api_url + "/terms/{}/canonicalized"
        elif decanonicalize:
            api_url = api_url + "/terms/{}/decanonicalized"
        else:
            log.warning("Missing (de)canonical flag - cannot convert namespaces")
            return nsarg
    else:

        api_url = (
            api_url + "/terms/{}/canonicalized"
        )  # overriding with namespace_targets

    request_url = api_url.format(url_path_param_quoting(nsarg))

    r = get_url(request_url, params=params, timeout=10)

    if r and r.status_code == 200:
        nsarg = r.json().get("term_id", nsarg)
    elif not r or r.status_code == 404:
        log.error(f"[de]Canonicalization endpoint missing: {request_url}")

    return nsarg


def convert_namespaces_str(
    bel_str: str,
    api_url: str = None,
    namespace_targets: Mapping[str, List[str]] = None,
    canonicalize: bool = False,
    decanonicalize: bool = False,
) -> str:
    """Convert namespace in string

    Uses a regex expression to extract all NSArgs and replace them with the
    updated NSArg from the BEL.bio API terms endpoint.

    Args:
        bel_str (str): bel statement string or partial string (e.g. subject or object)
        api_url (str): BEL.bio api url to use, e.g. https://api.bel.bio/v1
        namespace_targets (Mapping[str, List[str]]): formatted as in configuration file example
        canonicalize (bool): use canonicalize endpoint/namespace targets
        decanonicalize (bool): use decanonicalize endpoint/namespace targets

    Results:
        str: bel statement with namespaces converted
    """

    # pattern - look for capitalized namespace followed by colon
    #           and either a quoted string or a string that
    #           can include any char other than space, comma or ')'
    matches = re.findall(r'([A-Z]+:"(?:\\.|[^"\\])*"|[A-Z]+:(?:[^\),\s]+))', bel_str)
    for nsarg in matches:
        if "DEFAULT:" in nsarg:  # skip default namespaces
            continue

        updated_nsarg = convert_nsarg(
            nsarg,
            api_url=api_url,
            namespace_targets=namespace_targets,
            canonicalize=canonicalize,
            decanonicalize=decanonicalize,
        )
        if updated_nsarg != nsarg:
            bel_str = bel_str.replace(nsarg, updated_nsarg)

    return bel_str


def convert_namespaces_ast(
    ast,
    api_url: str = None,
    namespace_targets: Mapping[str, List[str]] = None,
    canonicalize: bool = False,
    decanonicalize: bool = False,
):
    """Recursively convert namespaces of BEL Entities in BEL AST using API endpoint

    Canonicalization and decanonicalization is determined by endpoint used and namespace_targets.

    Args:
        ast (BEL): BEL AST
        api_url (str): endpoint url with a placeholder for the term_id (either /terms/<term_id>/canonicalized or /terms/<term_id>/decanonicalized)
        namespace_targets (Mapping[str, List[str]]): (de)canonical targets for converting BEL Entities

    Returns:
        BEL: BEL AST
    """

    if isinstance(ast, NSArg):
        given_term_id = "{}:{}".format(ast.namespace, ast.value)

        # Get normalized term if necessary
        if (canonicalize and not ast.canonical) or (
            decanonicalize and not ast.decanonical
        ):
            normalized_term = convert_nsarg(
                given_term_id,
                api_url=api_url,
                namespace_targets=namespace_targets,
                canonicalize=canonicalize,
                decanonicalize=decanonicalize,
            )
            if canonicalize:
                ast.canonical = normalized_term
            elif decanonicalize:
                ast.decanonical = normalized_term

        # Update normalized term
        if canonicalize:
            ns, value = ast.canonical.split(":")
            ast.change_nsvalue(ns, value)
        elif decanonicalize:
            ns, value = ast.canonical.split(":")
            ast.change_nsvalue(ns, value)

    # Recursively process every NSArg by processing BELAst and Functions
    if hasattr(ast, "args"):
        for arg in ast.args:
            convert_namespaces_ast(
                arg,
                api_url=api_url,
                namespace_targets=namespace_targets,
                canonicalize=canonicalize,
                decanonicalize=decanonicalize,
            )

    return ast


def populate_ast_nsarg_defaults(ast, belast, species_id=None):
    """Recursively populate NSArg AST entries for default (de)canonical values

    This was added specifically for the BEL Pipeline. It is designed to
    run directly against ArangoDB and not through the BELAPI.

    Args:
        ast (BEL): BEL AST

    Returns:
        BEL: BEL AST
    """

    if isinstance(ast, NSArg):
        given_term_id = "{}:{}".format(ast.namespace, ast.value)

        r = bel.terms.terms.get_normalized_terms(given_term_id)
        ast.canonical = r["canonical"]
        ast.decanonical = r["decanonical"]

        r = bel.terms.terms.get_terms(ast.canonical)

        if len(r) > 0:
            ast.species_id = r[0].get("species_id", False)
            ast.species_label = r[0].get("species_label", False)

        # Check to see if species is set and if it's consistent
        #   if species is not consistent for the entire AST - set species_id/label
        #   on belast to False (instead of None)
        if ast.species_id and species_id is None:
            species_id = ast.species_id
            belast.species.add((ast.species_id, ast.species_label))

        elif ast.species_id and species_id and species_id != ast.species_id:
            belast.species_id = False
            belast.species_label = False

    # Recursively process every NSArg by processing BELAst and Functions
    if hasattr(ast, "args"):
        for arg in ast.args:
            populate_ast_nsarg_defaults(arg, belast, species_id)

    return ast


def orthologize(ast, bo, species_id: str):
    """Recursively orthologize BEL Entities in BEL AST using API endpoint

    NOTE: - will take first ortholog returned in BEL.bio API result (which may return more than one ortholog)

    Args:
        ast (BEL): BEL AST
        endpoint (str): endpoint url with a placeholder for the term_id

    Returns:
        BEL: BEL AST
    """

    # if species_id == 'TAX:9606' and str(ast) == 'MGI:Sult2a1':
    #     import pdb; pdb.set_trace()

    if not species_id:
        bo.validation_messages.append(
            ("WARNING", "No species id was provided for orthologization")
        )
        return ast

    if isinstance(ast, NSArg):
        if ast.orthologs:
            # log.debug(f'AST: {ast.to_string()}  species_id: {species_id}  orthologs: {ast.orthologs}')
            if ast.orthologs.get(species_id, None):
                orthologized_nsarg_val = ast.orthologs[species_id]["decanonical"]
                ns, value = orthologized_nsarg_val.split(":")
                ast.change_nsvalue(ns, value)
                ast.canonical = ast.orthologs[species_id]["canonical"]
                ast.decanonical = ast.orthologs[species_id]["decanonical"]
                ast.orthologized = True
                bo.ast.species.add(
                    (species_id, ast.orthologs[species_id]["species_label"])
                )
            else:
                bo.ast.species.add((ast.species_id, ast.species_label))
                bo.validation_messages.append(
                    ("WARNING", f"No ortholog found for {ast.namespace}:{ast.value}")
                )
        elif ast.species_id:
            bo.ast.species.add((ast.species_id, ast.species_label))

    # Recursively process every NSArg by processing BELAst and Functions
    if hasattr(ast, "args"):
        for arg in ast.args:
            orthologize(arg, bo, species_id)

    return ast


def populate_ast_nsarg_orthologs(ast, species):
    """Recursively collect NSArg orthologs for BEL AST

    This requires bo.collect_nsarg_norms() to be run first so NSArg.canonical is available

    Args:
        ast: AST at recursive point in belobj
        species: dictionary of species ids vs labels for or
    """

    ortholog_namespace = "EG"

    if isinstance(ast, NSArg):
        if re.match(ortholog_namespace, ast.canonical):
            orthologs = bel.terms.orthologs.get_orthologs(
                ast.canonical, list(species.keys())
            )
            for species_id in species:
                if species_id in orthologs:
                    orthologs[species_id]["species_label"] = species[species_id]

            ast.orthologs = copy.deepcopy(orthologs)

    # Recursively process every NSArg by processing BELAst and Functions
    if hasattr(ast, "args"):
        for arg in ast.args:
            populate_ast_nsarg_orthologs(arg, species)

    return ast


def preprocess_bel_stmt(stmt: str) -> str:
    """Clean up basic formatting of BEL statement

    Args:
        stmt: BEL statement as single string

    Returns:
        cleaned BEL statement
    """

    stmt = stmt.strip()  # remove newline at end of stmt
    stmt = re.sub(r",+", ",", stmt)  # remove multiple commas
    stmt = re.sub(r",", ", ", stmt)  # add space after each comma
    stmt = re.sub(r" +", " ", stmt)  # remove multiple spaces

    return stmt


# TODO remove AST normalize_nsarg_value for this and add tests
def quoting_nsarg(nsarg_value):
    """Quoting nsargs

    If needs quotes (only if it contains whitespace, comma or ')' ), make sure
        it is quoted, else don't add them.


    """
    quoted = re.findall(r'^"(.*)"$', nsarg_value)

    if re.search(
        r"[),\s]", nsarg_value
    ):  # quote only if it contains whitespace, comma or ')'
        if quoted:
            return nsarg_value
        else:
            return f'"{nsarg_value}"'
    else:
        if quoted:
            return quoted[0]
        else:
            return nsarg_value


# # See TODO in bel.py for this function - not currently enabled
# def simple_checks(stmt):
#     """Simple typo checks for BEL statement

#     Args:
#         stmt (str): BEL statement as single string

#     Returns:
#         Tuple[bool, List[Tuple[str, str]]]: is valid? and list of ...
#     """
#     messages = []
#     is_valid = True

#     # check for even number of parenthesis
#     left_p_ct = stmt.count('(')
#     right_p_ct = stmt.count(')')

#     if left_p_ct < right_p_ct:
#         messages.append(('ERROR', 'Unbalanced parenthesis: Missing left parenthesis somewhere!'))
#     elif right_p_ct < left_p_ct:
#         messages.append(('ERROR', 'Unbalanced parenthesis: Missing right parenthesis somewhere!'))

#     # check for even number of quotation marks
#     single_quote_ct = stmt.count('\'')
#     double_quote_ct = stmt.count('"')

#     if single_quote_ct > 0:  # single quotes not allowed
#         messages.append(('ERROR', 'Single quotes are not allowed! Please use double quotes.'))

#     if double_quote_ct % 2 != 0:  # odd number of quotations
#         messages.append(('ERROR', 'Unbalanced quotations: Missing quotation mark somewhere!'))

#     if messages:
#         is_valid = False

#     (is_valid, messages)


def handle_parser_syntax_error(e):
    col_failed = e.pos
    info = e.buf.line_info(e.pos)
    text = info.text.rstrip()
    leading = re.sub(r"[^\t]", " ", text)[: info.col]
    text = text.expandtabs()
    leading = leading.expandtabs()
    undefined_type = e.stack[-1]

    err_visualizer = "{}\n{}^".format(text, leading)
    if undefined_type == "relations":
        error_msg = (
            "Failed parse at position {}. "
            "Check that you have a valid relation.".format(col_failed)
        )
    elif undefined_type == "funcs":
        error_msg = (
            "Failed parse at position {}. "
            "Check that you have a valid primary or modifier function.".format(
                col_failed
            )
        )
    elif undefined_type == "function_open":
        error_msg = (
            "Failed parse at position {}. "
            "Check that you have have opened your parenthesis correctly before this point.".format(
                col_failed
            )
        )
    elif undefined_type == "function_close":
        error_msg = (
            "Failed parse at position {}. "
            "Check that you have have closed your parenthesis correctly before this point.".format(
                col_failed
            )
        )
    elif undefined_type == "full_nsv":
        error_msg = (
            "Failed parse at position {}. "
            "Check that you have a valid namespace argument.".format(col_failed)
        )
    else:
        error_msg = (
            "Failed parse at position {}. "
            "Check to make sure commas/spaces are not missing.".format(
                col_failed, undefined_type
            )
        )

    return error_msg, err_visualizer


def _dump_spec(spec):
    """Dump bel specification dictionary using YAML

    Formats this with an extra indentation for lists to make it easier to
    use cold folding on the YAML version of the spec dictionary.
    """
    with open("spec.yaml", "w") as f:
        yaml.dump(spec, f, Dumper=MyDumper, default_flow_style=False)


def _default_to_version(version, available_versions):

    if not available_versions:
        log.error("No versions available.")
        return None

    if any(char.isalpha() for char in version):
        log.error("Invalid version number entered. Examples: '2', '3.1', '3.2.6'.")
        return None

    version_semantic_regex = r"(\d+)(?:\.(\d+))?(?:\.(\d+))?"
    our_match = re.match(version_semantic_regex, version)

    if our_match:
        wanted_major = int(our_match.group(1)) if our_match.group(1) else "x"
        wanted_minor = int(our_match.group(2)) if our_match.group(2) else "x"
        wanted_patch = int(our_match.group(3)) if our_match.group(3) else "x"
        formatted_version = "{}.{}.{}".format(wanted_major, wanted_minor, wanted_patch)
    else:
        log.error("Invalid version number entered. Examples: '2', '3.1', '3.2.6'.")
        return None

    if formatted_version in available_versions:
        return formatted_version

    # now we need to find closest available version that is EQUAL OR GREATER

    available_versions.sort(key=lambda s: list(map(int, s.split("."))))

    best_choice = None

    for v in available_versions:
        v_split = v.split(".")
        v_maj = int(v_split[0])
        v_min = int(v_split[1])
        v_pat = int(v_split[2])

        if wanted_major == v_maj and wanted_minor == v_min and wanted_patch == v_pat:
            return v  # exact version found. return.
        elif wanted_major == v_maj and wanted_minor == v_min and wanted_patch == "x":
            best_choice = v  # continue to see if higher patch number available
            continue
        elif wanted_major == v_maj and wanted_minor == "x" and wanted_patch == "x":
            best_choice = v  # continue to see if higher minor/patch number available
            continue

    if best_choice is not None:
        log.error(
            "Version {} not available in library. Defaulting to {}.".format(
                version, best_choice
            )
        )
    else:
        log.error("Version {} not available in library.".format(version))

    return best_choice


class MyDumper(yaml.Dumper):
    def increase_indent(self, flow=False, indentless=False):
        return super(MyDumper, self).increase_indent(flow, False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Usage:  program.py <customer>

"""

from typing import Mapping, Any, List, MutableSequence
import copy
import json
import sys
import os

import bel.lang.belobj
import bel.lang.bel_specification
import bel.db.arangodb as arangodb
import bel.utils as utils


from bel.Config import config

import structlog

log = structlog.getLogger(__name__)

Edges = MutableSequence[Mapping[str, Any]]

arango_client = arangodb.get_client()
edgestore_db = arangodb.get_edgestore_handle(arango_client)


def nanopub_to_edges(nanopub: dict = {}, rules: List[str] = [], orthologize_targets: list = []):
    """Process nanopub into edges and load into EdgeStore

    Args:
        nanopub: BEL Nanopub
        rules: list of compute rules to process
        orthologize_targets: list of species in TAX:<int> format

    Returns:
        list: of edges

    Edge object:
        {
            "edge": {
                "subject": {
                    "name": subj_canon,
                    "name_lc": subj_canon.lower(),
                    "label": subj_lbl,
                    "label_lc": subj_lbl.lower(),
                    "components": subj_components,
                },
                "relation": {  # relation _key is based on a hash
                    "relation": edge_ast.bel_relation,
                    "edge_hash": edge_hash,
                    "edge_dt": edge_dt,
                    "nanopub_url": nanopub_url,
                    "nanopub_id": nanopub_id,
                    "citation": citation,
                    "subject_canon": subj_canon,
                    "subject": subj_lbl,
                    "object_canon": obj_canon,
                    "object": obj_lbl,
                    "annotations": nanopub['annotations'],
                    "metadata": nanopub['metadata'],
                    "public_flag": True,  # will be added when groups/permissions feature is finished,
                    "edge_types": edge_types,
                },
                'object': {
                    "name": obj_canon,
                    "name_lc": obj_canon.lower(),
                    "label": obj_lbl,
                    "label_lc": obj_lbl.lower(),
                    "components": obj_components,
                }
            }
        }
    """

    # Collect input values ####################################################
    nanopub_url = nanopub.get("source_url", "")

    edge_dt = utils.dt_utc_formatted()  # don't want this in relation_id

    # Extract BEL Version and make sure we can process this
    if nanopub["nanopub"]["type"]["name"].upper() == "BEL":
        bel_version = nanopub["nanopub"]["type"]["version"]
        versions = bel.lang.bel_specification.get_bel_versions()
        if bel_version not in versions:
            log.error(
                f"Do not know this BEL Version: {bel_version}, these are the ones I can process: {versions.keys()}"
            )
            return []
    else:
        log.error(
            f"Not a BEL Nanopub according to nanopub.type.name: {nanopub['nanopub']['type']['name']}"
        )
        return []

    # Required for BEL parsing/canonicalization/orthologization
    api_url = config["bel_api"]["servers"]["api_url"]

    try:
        citation_string = normalize_nanopub_citation(nanopub)
    except Exception as e:
        log.error(f"Could not create citation string for {nanopub_url}")
        citation_string = ""

    if orthologize_targets == []:
        if config["bel_api"].get("edges", None):
            orthologize_targets = config["bel_api"]["edges"].get("orthologize_targets", [])

    # orig_species_id = [anno['id'] for anno in nanopub['nanopub']['annotations'] if anno['type'] == 'Species']
    # if orig_species_id:
    #     orig_species_id = orig_species_id[0]

    master_annotations = copy.deepcopy(nanopub["nanopub"]["annotations"])
    master_metadata = copy.deepcopy(nanopub["nanopub"]["metadata"])
    master_metadata.pop("gd_abstract", None)

    nanopub_type = nanopub["nanopub"]["metadata"].get("nanopub_type")

    # Create Edge Assertion Info ##############################################
    # r = generate_assertion_edge_info(nanopub['nanopub']['assertions'], orig_species_id, orthologize_targets, bel_version, api_url, nanopub_type)
    r = generate_assertion_edge_info(
        nanopub["nanopub"]["assertions"], orthologize_targets, bel_version, api_url, nanopub_type
    )
    edge_info_list = r["edge_info_list"]

    # Build Edges #############################################################
    edges = []
    errors = []
    for edge_info in edge_info_list:
        annotations = copy.deepcopy(master_annotations)
        metadata = copy.deepcopy(master_metadata)

        errors.extend(edge_info["errors"])

        if not edge_info.get("canonical"):
            continue

        # TODO - remove this
        # if edge_info.get('species_id', False):
        #     annotations = orthologize_context(edge_info['species_id'], annotations)

        edge_hash = utils._create_hash(
            f'{edge_info["canonical"]["subject"]} {edge_info["canonical"]["relation"]} {edge_info["canonical"]["object"]}'
        )

        edge = {
            "edge": {
                "subject": {
                    "name": edge_info["canonical"]["subject"],
                    "name_lc": edge_info["canonical"]["subject"].lower(),
                    "label": edge_info["decanonical"]["subject"],
                    "label_lc": edge_info["decanonical"]["subject"].lower(),
                    "components": edge_info["subject_comp"],
                },
                "relation": {
                    "relation": edge_info["canonical"]["relation"],
                    "edge_hash": edge_hash,
                    "edge_dt": edge_dt,
                    "nanopub_url": nanopub_url,
                    "nanopub_id": nanopub["nanopub"]["id"],
                    "citation": citation_string,
                    "subject_canon": edge_info["canonical"]["subject"],
                    "subject": edge_info["decanonical"]["subject"],
                    "object_canon": edge_info["canonical"]["object"],
                    "object": edge_info["decanonical"]["object"],
                    "annotations": copy.deepcopy(annotations),
                    "metadata": copy.deepcopy(metadata),
                    "public_flag": True,
                    "edge_types": edge_info["edge_types"],
                    "species_id": edge_info["species_id"],
                    "species_label": edge_info["species_label"],
                },
                "object": {
                    "name": edge_info["canonical"]["object"],
                    "name_lc": edge_info["canonical"]["object"].lower(),
                    "label": edge_info["decanonical"]["object"],
                    "label_lc": edge_info["decanonical"]["object"].lower(),
                    "components": edge_info["object_comp"],
                },
            }
        }

        edges.append(copy.deepcopy(edge))

    return {
        "edges": edges,
        "nanopub_id": nanopub["nanopub"]["id"],
        "nanopub_url": nanopub_url,
        "success": True,
        "errors": errors,
    }


def extract_ast_species(ast):
    """Extract species from ast.species set of tuples (id, label)"""

    species_id = "None"
    species_label = "None"
    species = [
        (species_id, species_label) for (species_id, species_label) in ast.species if species_id
    ]

    if len(species) == 1:
        (species_id, species_label) = species[0]

    if not species_id:
        species_id = "None"
        species_label = "None"

    log.debug(f"AST Species: {ast.species}  Species: {species}  SpeciesID: {species_id}")

    return (species_id, species_label)


# def generate_assertion_edge_info(assertions: List[dict], orig_species_id: str, orthologize_targets: List[str], bel_version: str, api_url: str, nanopub_type: str) -> dict:
def generate_assertion_edge_info(
    assertions: List[dict],
    orthologize_targets: List[str],
    bel_version: str,
    api_url: str,
    nanopub_type: str = "",
) -> dict:
    """Create edges (SRO) for assertions given orthologization targets

    Args:
        assertions: list of BEL statements (SRO object)
        orthologize_targets: list of species in TAX:<int> format
        bel_version: to be used for processing assertions
        api_url: BEL API url endpoint to use for terminologies and orthologies
    """

    bo = bel.lang.belobj.BEL(bel_version, api_url)
    bo_computed = bel.lang.belobj.BEL(bel_version, api_url)

    edge_info_list = []

    with utils.Timer() as t:
        for assertion in assertions:
            # if not assertion.get('relation', False):
            #     continue  # Skip any subject only statements

            start_time = t.elapsed
            bo.parse(assertion)

            if not bo.ast:
                errors = [
                    f"{error[0]} {error[1]}"
                    for error in bo.validation_messages
                    if error[0] == "ERROR"
                ]
                edge_info = {"errors": copy.deepcopy(errors)}
                edge_info_list.append(copy.deepcopy(edge_info))
                continue

            # populate canonical terms and orthologs for assertion
            bo.collect_nsarg_norms()
            bo.collect_orthologs(orthologize_targets)

            log.info(
                "Timing - time to collect nsargs and orthologs", delta_ms=(t.elapsed - start_time)
            )

            (edge_species_id, edge_species_label) = extract_ast_species(bo.ast)
            orig_species_id = edge_species_id

            canon = bo.canonicalize().to_triple()

            components = get_node_subcomponents(bo.ast)  # needs to be run after canonicalization
            computed_asts = bo.compute_edges(
                ast_result=True
            )  # needs to be run after canonicalization
            decanon = bo.decanonicalize().to_triple()

            if nanopub_type == "backbone":
                edge_types = ["backbone"]
            else:
                edge_types = ["original", "primary"]

            if assertion.get("relation", False):

                causal_edge_type = []
                if "causal" in bo.spec["relations"]["info"][assertion["relation"]]["categories"]:
                    edge_types.append("causal")

                edge_info = {
                    "edge_types": edge_types,
                    "species_id": edge_species_id,
                    "species_label": edge_species_label,
                    "canonical": canon,
                    "decanonical": decanon,
                    "subject_comp": components["subject_comp"],
                    "object_comp": components["object_comp"],
                    "errors": [],
                }
                edge_info_list.append(copy.deepcopy(edge_info))

            # Loop through primary computed asts
            for computed_ast in computed_asts:
                bo_computed.ast = computed_ast
                bo_computed.collect_nsarg_norms()
                canon = bo_computed.canonicalize().to_triple()
                components = get_node_subcomponents(
                    bo_computed.ast
                )  # needs to be run after canonicalization
                decanon = bo_computed.decanonicalize().to_triple()

                (edge_species_id, edge_species_label) = extract_ast_species(bo_computed.ast)

                edge_info = {
                    "edge_types": ["computed"],
                    "species_id": edge_species_id,
                    "species_label": edge_species_label,
                    "canonical": canon,
                    "decanonical": decanon,
                    "subject_comp": components["subject_comp"],
                    "object_comp": components["object_comp"],
                    "errors": [],
                }
                if [edge for edge in edge_info_list if edge.get("canonical", {}) == canon]:
                    continue  # skip if edge is already included (i.e. the primary is same as computed edge)
                edge_info_list.append(copy.deepcopy(edge_info))

            # Skip orthologs if backbone nanopub
            if nanopub_type == "backbone":
                continue

            # only process orthologs if there are species-specific NSArgs
            if len(bo.ast.species) > 0:
                # Loop through orthologs
                for species_id in orthologize_targets:
                    log.debug(f"Orig species: {orig_species_id}  Target species: {species_id}")
                    if species_id == orig_species_id:
                        continue

                    bo.orthologize(species_id)

                    (edge_species_id, edge_species_label) = extract_ast_species(bo.ast)

                    if edge_species_id == "None" or edge_species_id == orig_species_id:
                        log.debug(
                            f'Skipping orthologization- species == "None" or {orig_species_id}  ASTspecies: {bo.ast.species} for {bo}'
                        )
                        continue

                    ortho_decanon = bo.orthologize(
                        species_id
                    ).to_triple()  # defaults to decanonicalized orthologized form
                    ortho_canon = bo.canonicalize().to_triple()
                    computed_asts = bo.compute_edges(
                        ast_result=True
                    )  # needs to be run after canonicalization
                    components = get_node_subcomponents(
                        bo.ast
                    )  # needs to be run after canonicalization

                    if assertion.get("relation", False):
                        edge_info = {
                            "edge_types": ["orthologized", "primary"] + causal_edge_type,
                            "species_id": edge_species_id,
                            "species_label": edge_species_label,
                            "canonical": ortho_canon,
                            "decanonical": ortho_decanon,
                            "subject_comp": components["subject_comp"],
                            "object_comp": components["object_comp"],
                            "errors": [],
                        }

                        edge_info_list.append(copy.deepcopy(edge_info))

                    # Loop through orthologized computed asts
                    for computed_ast in computed_asts:
                        bo_computed.ast = computed_ast
                        bo_computed.collect_nsarg_norms()
                        canon = bo_computed.canonicalize().to_triple()
                        components = get_node_subcomponents(
                            bo_computed.ast
                        )  # needs to be run after canonicalization
                        decanon = bo_computed.decanonicalize().to_triple()

                        (edge_species_id, edge_species_label) = extract_ast_species(bo_computed.ast)

                        edge_info = {
                            "edge_types": ["computed", "orthologized"],
                            "species_id": edge_species_id,
                            "species_label": edge_species_label,
                            "canonical": canon,
                            "decanonical": decanon,
                            "subject_comp": components["subject_comp"],
                            "object_comp": components["object_comp"],
                            "errors": [],
                        }
                        if [edge for edge in edge_info_list if edge.get("canonical", {}) == canon]:
                            continue  # skip if edge is already included (i.e. the primary is same as computed edge)
                        edge_info_list.append(copy.deepcopy(edge_info))

    log.info("Timing - Generated all edge info for all nanopub assertions", delta_ms=t.elapsed)

    return {"edge_info_list": edge_info_list}


def get_node_subcomponents(ast):

    sub, obj = [], []

    try:
        sub = ast.bel_subject.subcomponents(subcomponents=[])

        # TODO - update handling nested BEL statement - see Natalie's recommendation on how to handle
        obj_components = []
        if ast.bel_object.__class__.__name__ == "BELAst":  # Nested BEL Assertion
            obj_components = ast.bel_object.bel_subject.subcomponents(subcomponents=[])
            obj_components = ast.bel_object.bel_object.subcomponents(subcomponents=obj_components)
        elif hasattr(ast.bel_object, "subcomponents"):  # Normal BEL Assertion
            obj_components = ast.bel_object.subcomponents(subcomponents=[])

        obj = obj_components

    except Exception as e:
        log.warning(f"Problem getting subcomponents for {ast.to_string()}", error=str(e))

    return {"subject_comp": sub, "object_comp": obj}


def orthologize_context(
    orthologize_target: str, annotations: Mapping[str, Any]
) -> Mapping[str, Any]:
    """Orthologize context

    Replace Species context with new orthologize target and add a annotation type of OrthologizedFrom
    """

    url = f'{config["bel_api"]["servers"]["api_url"]}/terms/{orthologize_target}'
    r = utils.get_url(url)
    species_label = r.json().get("label", "unlabeled")

    orthologized_from = {}
    for idx, annotation in enumerate(annotations):
        if annotation["type"] == "Species":
            orthologized_from = {"id": annotation["id"], "label": annotation["label"]}
            annotations[idx] = {"type": "Species", "id": orthologize_target, "label": species_label}

    if "id" in orthologized_from:
        annotations.append(
            {
                "type": "OrigSpecies",
                "id": f'Orig-{orthologized_from["id"]}',
                "label": f'Orig-{orthologized_from["label"]}',
            }
        )

    return annotations


def normalize_nanopub_citation(nanopub):

    citation_string = ""
    if "database" in nanopub["nanopub"]["citation"]:
        if nanopub["nanopub"]["citation"]["database"]["name"].lower() == "pubmed":
            citation_string = f"PMID:{nanopub['nanopub']['citation']['database']['id']}"
        else:
            citation_string = f"{nanopub['nanopub']['citation']['database']['name']}:{nanopub['nanopub']['citation']['database']['id']}"
    elif "reference" in nanopub["nanopub"]["citation"] and isinstance(
        nanopub["nanopub"]["citation"]["reference"], str
    ):
        citation_string = nanopub["nanopub"]["citation"]["reference"]
    elif "uri" in nanopub["nanopub"]["citation"]:
        citation_string = nanopub["nanopub"]["citation"]["uri"]

    return citation_string


def main():
    pass


if __name__ == "__main__":
    # Setup logging
    import logging.config

    module_fn = os.path.basename(__file__)
    module_fn = module_fn.replace(".py", "")

    if config.get("logging", False):
        logging.config.dictConfig(config.get("logging"))

    log = logging.getLogger(f"{module_fn}")

    main()



from typing import List

import bel.db.arangodb
import bel.terms.terms

import structlog

log = structlog.getLogger()

default_canonical_namespace = "EG"  # for genes, proteins

arangodb_client = bel.db.arangodb.get_client()
belns_db = bel.db.arangodb.get_belns_handle(arangodb_client)


def get_orthologs(canonical_gene_id: str, species: list = []) -> List[dict]:
    """Get orthologs for given gene_id and species

    Canonicalize prior to ortholog query and decanonicalize
    the resulting ortholog

    Args:
        canonical_gene_id: canonical gene_id for which to retrieve ortholog
        species: target species for ortholog - tax id format TAX:<number>

    Returns:
        List[dict]: {'tax_id': <tax_id>, 'canonical': canonical_id, 'decanonical': decanonical_id}
    """

    gene_id_key = bel.db.arangodb.arango_id_to_key(canonical_gene_id)
    orthologs = {}

    if species:
        query_filter = f"FILTER vertex.tax_id IN {species}"

    query = f"""
        LET start = (
            FOR vertex in ortholog_nodes
                FILTER vertex._key == "{gene_id_key}"
                RETURN {{ "name": vertex.name, "tax_id": vertex.tax_id }}
        )

        LET orthologs = (
            FOR vertex IN 1..3
                ANY "ortholog_nodes/{gene_id_key}" ortholog_edges
                OPTIONS {{ bfs: true, uniqueVertices : 'global' }}
                {query_filter}
                RETURN DISTINCT {{ "name": vertex.name, "tax_id": vertex.tax_id }}
        )

        RETURN {{ 'orthologs': FLATTEN(UNION(start, orthologs)) }}
    """

    cursor = belns_db.aql.execute(query, batch_size=20)
    results = cursor.pop()
    for ortholog in results["orthologs"]:
        norms = bel.terms.terms.get_normalized_terms(ortholog["name"])
        orthologs[ortholog["tax_id"]] = {
            "canonical": norms["canonical"],
            "decanonical": norms["decanonical"],
        }

    return orthologs



import timy
import json
import gzip
import copy

from arango import ArangoError

from typing import IO

import bel.utils
import bel.db.elasticsearch as elasticsearch
import bel.db.arangodb as arangodb

from bel.Config import config

# import structlog
# import logging
# log = logging.getLogger(__name__)

from structlog import get_logger

log = get_logger()

terms_alias = "terms"


def load_terms(fo: IO, metadata: dict, forceupdate: bool):
    """Load terms into Elasticsearch and ArangoDB

    Forceupdate will create a new index in Elasticsearch regardless of whether
    an index with the resource version already exists.

    Args:
        fo: file obj - terminology file
        metadata: dict containing the metadata for terminology
        forceupdate: force full update - e.g. don't leave Elasticsearch indexes
            alone if their version ID matches
    """

    version = metadata["metadata"]["version"]

    # LOAD TERMS INTO Elasticsearch
    with timy.Timer("Load Terms") as timer:
        es = bel.db.elasticsearch.get_client()

        es_version = version.replace("T", "").replace("-", "").replace(":", "")
        index_prefix = f"terms_{metadata['metadata']['namespace'].lower()}"
        index_name = f"{index_prefix}_{es_version}"

        # Create index with mapping
        if not elasticsearch.index_exists(es, index_name):
            elasticsearch.create_terms_index(es, index_name)
        elif forceupdate:  # force an update to the index
            index_name += "_alt"
            elasticsearch.create_terms_index(es, index_name)
        else:
            return  # Skip loading if not forced and not a new namespace

        terms_iterator = terms_iterator_for_elasticsearch(fo, index_name)
        elasticsearch.bulk_load_docs(es, terms_iterator)

        # Remove old namespace index
        index_names = elasticsearch.get_all_index_names(es)
        for name in index_names:
            if name != index_name and index_prefix in name:
                elasticsearch.delete_index(es, name)

        # Add terms_alias to this index
        elasticsearch.add_index_alias(es, index_name, terms_alias)

        log.info(
            "Load namespace terms",
            elapsed=timer.elapsed,
            namespace=metadata["metadata"]["namespace"],
        )

    # LOAD EQUIVALENCES INTO ArangoDB
    with timy.Timer("Load Term Equivalences") as timer:
        arango_client = arangodb.get_client()
        belns_db = arangodb.get_belns_handle(arango_client)
        arangodb.batch_load_docs(
            belns_db, terms_iterator_for_arangodb(fo, version), on_duplicate="update"
        )

        log.info(
            "Loaded namespace equivalences",
            elapsed=timer.elapsed,
            namespace=metadata["metadata"]["namespace"],
        )

        # Clean up old entries
        remove_old_equivalence_edges = f"""
            FOR edge in equivalence_edges
                FILTER edge.source == "{metadata["metadata"]["namespace"]}"
                FILTER edge.version != "{version}"
                REMOVE edge IN equivalence_edges
        """
        remove_old_equivalence_nodes = f"""
            FOR node in equivalence_nodes
                FILTER node.source == "{metadata["metadata"]["namespace"]}"
                FILTER node.version != "{version}"
                REMOVE node IN equivalence_nodes
        """
        arangodb.aql_query(belns_db, remove_old_equivalence_edges)
        arangodb.aql_query(belns_db, remove_old_equivalence_nodes)

    # Add metadata to resource metadata collection
    metadata["_key"] = f"Namespace_{metadata['metadata']['namespace']}"
    try:
        belns_db.collection(arangodb.belns_metadata_name).insert(metadata)
    except ArangoError as ae:
        belns_db.collection(arangodb.belns_metadata_name).replace(metadata)


def terms_iterator_for_arangodb(fo, version):

    species_list = config["bel_resources"].get("species_list", [])

    fo.seek(0)
    with gzip.open(fo, "rt") as f:
        for line in f:
            term = json.loads(line)
            # skip if not term record (e.g. is a metadata record)
            if "term" not in term:
                continue
            term = term["term"]

            species_id = term.get("species_id", None)
            # Skip if species not listed in species_list
            if species_list and species_id and species_id not in species_list:
                continue

            source = term["namespace"]
            term_id = term["id"]
            term_key = arangodb.arango_id_to_key(term_id)

            (ns, val) = term_id.split(":", maxsplit=1)

            # Add primary ID node
            yield (
                arangodb.equiv_nodes_name,
                {
                    "_key": term_key,
                    "name": term_id,
                    "primary": True,
                    "namespace": ns,
                    "source": source,
                    "version": version,
                },
            )

            # Create Alt ID nodes/equivalences (to support other database equivalences using non-preferred Namespace IDs)
            if "alt_ids" in term:
                for alt_id in term["alt_ids"]:
                    # log.info(f'Added {alt_id} equivalence')
                    alt_id_key = arangodb.arango_id_to_key(alt_id)
                    yield (
                        arangodb.equiv_nodes_name,
                        {
                            "_key": alt_id_key,
                            "name": alt_id,
                            "namespace": ns,
                            "source": source,
                            "version": version,
                        },
                    )

                    arango_edge = {
                        "_from": f"{arangodb.equiv_nodes_name}/{term_key}",
                        "_to": f"{arangodb.equiv_nodes_name}/{alt_id_key}",
                        "_key": bel.utils._create_hash(f"{term_id}>>{alt_id}"),
                        "type": "equivalent_to",
                        "source": source,
                        "version": version,
                    }
                    yield (arangodb.equiv_edges_name, arango_edge)

            # Cross-DB equivalences
            if "equivalences" in term:
                for eqv in term["equivalences"]:
                    (ns, val) = eqv.split(":", maxsplit=1)
                    eqv_key = arangodb.arango_id_to_key(eqv)

                    yield (
                        arangodb.equiv_nodes_name,
                        {
                            "_key": eqv_key,
                            "name": eqv,
                            "namespace": ns,
                            "source": source,
                            "version": version,
                        },
                    )

                    arango_edge = {
                        "_from": f"{arangodb.equiv_nodes_name}/{term_key}",
                        "_to": f"{arangodb.equiv_nodes_name}/{eqv_key}",
                        "_key": bel.utils._create_hash(f"{term_id}>>{eqv}"),
                        "type": "equivalent_to",
                        "source": source,
                        "version": version,
                    }
                    yield (arangodb.equiv_edges_name, arango_edge)


def terms_iterator_for_elasticsearch(fo: IO, index_name: str):
    """Add index_name to term documents for bulk load"""

    species_list = config["bel_resources"].get("species_list", [])

    fo.seek(0)  # Seek back to beginning of file
    with gzip.open(fo, "rt") as f:
        for line in f:
            term = json.loads(line)
            # skip if not term record (e.g. is a metadata record)
            if "term" not in term:
                continue
            term = term["term"]

            # Filter species if enabled in config
            species_id = term.get("species_id", None)
            if species_list and species_id and species_id not in species_list:
                continue

            all_term_ids = set()
            for term_id in [term["id"]] + term.get("alt_ids", []):
                all_term_ids.add(term_id)
                all_term_ids.add(lowercase_term_id(term_id))

            term["alt_ids"] = copy.copy(list(all_term_ids))

            yield {
                "_op_type": "index",
                "_index": index_name,
                "_type": "term",
                "_id": term["id"],
                "_source": copy.deepcopy(term),
            }


def lowercase_term_id(term_id: str) -> str:
    """Lowercase the term value (not the namespace prefix)

    Args:
        term_id (str): term identifier with namespace prefix, e.g. MESH:Atherosclerosis

    Returns:
        str: lowercased, e.g. MESH:atherosclerosis
    """
    (ns, val) = term_id.split(":", maxsplit=1)
    term_id = f"{ns}:{val.lower()}"

    return term_id



"""Various utilities used throughout the BEL package

NOTE: reqsess allows caching external (including ElasticSearch and ArangoDB) REST
      requests for 1 day. This can cause stale results to show up. The cache
      lives for the life of the application using *bel* so you'll need to restart
      the BEL.bio API server if you update the terminologies, you expect major
      changes in the Pubtator results, etc.
"""

import ulid
import tempfile
from cityhash import CityHash64
import json
from typing import Mapping, Any
import datetime
import dateutil
import requests
import requests_cache

from structlog import get_logger

log = get_logger()

requests_cache.install_cache(backend="sqlite", expire_after=600)


def get_url(url: str, params: dict = {}, timeout: float = 5.0, cache: bool = True):
    """Wrapper for requests.get(url)

    Args:
        url: url to retrieve
        params: query string parameters
        timeout: allow this much time for the request and time it out if over
        cache: Cache for up to a day unless this is false

    Returns:
        Requests Result obj or None if timed out
    """

    try:

        if not cache:
            with requests_cache.disabled():
                r = requests.get(url, params=params, timeout=timeout)
        else:
            r = requests.get(url, params=params, timeout=timeout)

        log.debug(f"Response headers {r.headers}  From cache {r.from_cache}")
        return r

    except requests.exceptions.Timeout:
        log.warn(f"Timed out getting url in get_url: {url}")
        return None
    except Exception as e:
        log.warn(f"Error getting url: {url}  error: {e}")
        return None


def timespan(start_time):
    """Return time in milliseconds from start_time"""

    timespan = datetime.datetime.now() - start_time
    timespan_ms = timespan.total_seconds() * 1000
    return timespan_ms


def download_file(url):
    """Download file"""

    response = requests.get(url, stream=True)
    fp = tempfile.NamedTemporaryFile()
    for chunk in response.iter_content(chunk_size=1024):
        if chunk:  # filter out keep-alive new chunks
            fp.write(chunk)

    # log.info(f'Download file - tmp file: {fp.name}  size: {fp.tell()}')
    return fp


def url_path_param_quoting(param):
    """Quote URL path parameters

    Convert '/' to _FORWARDSLASH_ - otherwise is interpreted as additional path parameter
        gunicorn processes the path prior to Falcon and interprets the
        correct quoting of %2F into a slash
    """
    return param.replace("/", "_FORWARDSLASH_")


def first_true(iterable, default=False, pred=None):
    """Returns the first true value in the iterable.

    If no true value is found, returns *default*

    If *pred* is not None, returns the first item
    for which pred(item) is true.

    """
    # first_true([a,b,c], x) --> a or b or c or x
    # first_true([a,b], x, f) --> a if f(a) else b if f(b) else x
    return next(filter(pred, iterable), default)


def _create_hash_from_doc(doc: Mapping[str, Any]) -> str:
    """Create hash Id from edge record

    Args:
        edge (Mapping[str, Any]): edge record to create hash from

    Returns:
        str: Murmur3 128 bit hash
    """

    doc_string = json.dumps(doc, sort_keys=True)
    return _create_hash(doc_string)


def _create_hash(string: str) -> str:
    """Create CityHash64 bit hash of string

    Args:
        string (str): string to create CityHash64 from

    Returns:
        str: CityHash64
    """

    return str(CityHash64(string))


def _generate_id() -> str:
    """Create ULID

    See: https://github.com/ahawker/ulid

    Returns:
        str: ULID random, unique identifier
    """

    # return str(uuid1())
    return ulid.new()


def dt_utc_formatted():
    """Create UTC ISODate formatted datetime string

    Format: YYYY-MM-DDThh:mm:ss.sssZ
    """
    return f"{datetime.datetime.utcnow().isoformat(timespec='milliseconds')}Z"


def parse_dt(dt: str):
    """Get datetime object from datetime strings"""

    return dateutil.parse(dt)


"""
https://github.com/brouberol/contexttimer/blob/master/contexttimer/__init__.py

Ctimer - A timer context manager measuring the
clock wall time of the code block it contains.
Copyright (C) 2013 Balthazar Rouberol - <brouberol@imap.cc>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__version__ = "0.3.3"


import functools
import collections
import logging

from timeit import default_timer


class Timer(object):

    """ A timer as a context manager
    Wraps around a timer. A custom timer can be passed
    to the constructor. The default timer is timeit.default_timer.
    Note that the latter measures wall clock time, not CPU time!
    On Unix systems, it corresponds to time.time.
    On Windows systems, it corresponds to time.clock.
    Keyword arguments:
        output -- if True, print output after exiting context.
                  if callable, pass output to callable.
        factor -- 1000 for milliseconds, 1 for seconds
        format -- str.format string to be used for output; default "took {} ms"
        prefix -- string to prepend (plus a space) to output
                  For convenience, if you only specify this, output defaults to True.
    """

    def __init__(
        self,
        timer=default_timer,
        factor=1000,
        output=None,
        fmt="took {:.1f} ms",
        prefix="",
    ):
        self.timer = timer
        self.factor = factor
        self.output = output
        self.fmt = fmt
        self.prefix = prefix
        self.end = None

    def __call__(self):
        """ Return the current time """
        return self.timer()

    def __enter__(self):
        """ Set the start time """
        self.start = self()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """ Set the end time """
        self.end = self()

        if self.prefix and self.output is None:
            self.output = True

        if self.output:
            output = " ".join([self.prefix, self.fmt.format(self.elapsed)])
            if callable(self.output):
                self.output(output)
            else:
                print(output)

    def __str__(self):
        return "%.3f" % (self.elapsed)

    @property
    def elapsed(self):
        """ Return the current elapsed time since start
        If the `elapsed` property is called in the context manager scope,
        the elapsed time bewteen start and property access is returned.
        However, if it is accessed outside of the context manager scope,
        it returns the elapsed time bewteen entering and exiting the scope.
        The `elapsed` property can thus be accessed at different points within
        the context manager scope, to time different parts of the block.
        """
        if self.end is None:
            # if elapsed is called in the context manager scope
            return (self() - self.start) * self.factor
        else:
            # if elapsed is called out of the context manager scope
            return (self.end - self.start) * self.factor


def timer(
    logger=None,
    level=logging.INFO,
    fmt="function %(function_name)s execution time: %(execution_time).3f",
    *func_or_func_args,
    **timer_kwargs,
):
    """ Function decorator displaying the function execution time
    All kwargs are the arguments taken by the Timer class constructor.
    """
    # store Timer kwargs in local variable so the namespace isn't polluted
    # by different level args and kwargs

    def wrapped_f(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            with Timer(**timer_kwargs) as t:
                out = f(*args, **kwargs)
            context = {"function_name": f.__name__, "execution_time": t.elapsed}
            if logger:
                logger.log(level, fmt % context, extra=context)
            else:
                print(fmt % context)
            return out

        return wrapped

    if len(func_or_func_args) == 1 and isinstance(
        func_or_func_args[0], collections.Callable
    ):
        return wrapped_f(func_or_func_args[0])
    else:
        return wrapped_f



#!/usr/bin/env python3.6
# -*- coding: utf-8 -*-

"""
Process belscripts content into nanopubs_bel-1.0.0 format for use in BEL.bio

BEL Script format documented here:

    https://wiki.openbel.org/display/BFD/BEL+Script+V2.0+Format

Notes:
    * Don't check that document and definition sections are at top of doc
    *
"""

import re
import json
import yaml
import copy
import gzip
import collections
import sys
import csv

import structlog

log = structlog.getLogger(__name__)

# citation fields are document type, a document name, a document reference ID, and an optional publication date, authors list and comment field

nanopub_type = {"name": "BEL", "version": "2.0.0"}


def convert_csv_str_to_list(csv_str: str) -> list:
    """Convert CSV str to list"""

    csv_str = re.sub("^\s*{", "", csv_str)
    csv_str = re.sub("}\s*$", "", csv_str)
    r = csv.reader([csv_str])
    row = list(r)[0]
    new = []
    for col in row:
        col = re.sub('^\s*"?\s*', "", col)
        col = re.sub('\s*"?\s*$', "", col)
        new.append(col)

    return new


def process_citation(citation_str: str) -> dict:
    """Parse BEL Script Citation string into nanopub_bel citation object"""

    citation_obj = {}

    citation_list = convert_csv_str_to_list(citation_str)
    (citation_type, name, doc_id, pub_date, authors, comment, *extra) = (
        citation_list + [None] * 7
    )
    # print(f'citation_type: {citation_type}, name: {name}, doc_id: {doc_id}, pub_date: {pub_date}, authors: {authors}, comment: {comment}')

    authors_list = []
    if authors:
        authors_list = authors.split("|")
        citation_obj["authors"] = authors_list

    if name and re.match("http?://", name):
        citation_obj["uri"] = name

    elif citation_type and citation_type.upper() == "PUBMED":
        citation_obj["database"] = {"name": "PubMed", "id": doc_id}
        if name:
            citation_obj["reference"] = name
    elif name:
        citation_obj["reference"] = name

    else:
        citation_obj["reference"] = "No reference found"

    if pub_date:
        citation_obj["date_published"] = pub_date

    if comment:
        citation_obj["comment"] = comment

    return citation_obj


def yield_metadata(nanopubs_metadata):
    """Yield nanopub metadata collected from BEL Script"""

    return {"metadata": copy.deepcopy(nanopubs_metadata)}


def split_bel_stmt(stmt: str, line_num) -> tuple:
    """Split bel statement into subject, relation, object tuple"""

    m = re.match(f"^(.*?\))\s+([a-zA-Z=\->\|:]+)\s+([\w(]+.*?)$", stmt, flags=0)
    if m:
        return (m.group(1), m.group(2), m.group(3))
    else:
        log.info(
            f"Could not parse bel statement into components at line number: {line_num} assertion: {stmt}"
        )
        return (stmt, None, None)


def yield_nanopub(assertions, annotations, line_num):
    """Yield nanopub object"""

    if not assertions:
        return {}

    anno = copy.deepcopy(annotations)

    evidence = anno.pop("evidence", None)
    stmt_group = anno.pop("statement_group", None)
    citation = anno.pop("citation", None)

    anno_list = []
    for anno_type in anno:
        if isinstance(anno[anno_type], (list, tuple)):
            for val in anno[anno_type]:
                anno_list.append({"type": anno_type, "label": val})
        else:
            anno_list.append({"type": anno_type, "label": anno[anno_type]})

    assertions_list = []
    for assertion in assertions:
        (subj, rel, obj) = split_bel_stmt(assertion, line_num)
        assertions_list.append({"subject": subj, "relation": rel, "object": obj})

    nanopub = {
        "schema_uri": "https://raw.githubusercontent.com/belbio/schemas/master/schemas/nanopub_bel-1.0.0.yaml",
        "type": copy.deepcopy(nanopub_type),
        "annotations": copy.deepcopy(anno_list),
        "citation": copy.deepcopy(citation),
        "assertions": copy.deepcopy(assertions_list),
        "evidence": evidence,
        "metadata": {"statement_group": stmt_group},
    }

    return {"nanopub": copy.deepcopy(nanopub)}


def process_documentline(line, nanopubs_metadata):
    """Process SET DOCUMENT line in BEL script"""

    matches = re.match('SET DOCUMENT\s+(\w+)\s+=\s+"?(.*?)"?$', line)
    key = matches.group(1)
    val = matches.group(2)
    nanopubs_metadata[key] = val

    return nanopubs_metadata


def process_definition(line, nanopubs_metadata):
    """Process DEFINE line in BEL script"""

    matches = re.match('DEFINE\s+(\w+)\s+(\w+)\s+AS\s+URL\s+"(.*?)"\s*$', line)
    if matches:
        def_type = matches.group(1).lower()
        if def_type == "namespace":
            def_type = "namespaces"
        elif def_type == "annotation":
            def_type == "annotations"

        key = matches.group(2)
        val = matches.group(3)

        if def_type in nanopubs_metadata:
            nanopubs_metadata[def_type][key] = val
        else:
            nanopubs_metadata[def_type] = {key: val}

    matches = re.match("DEFINE\s+(\w+)\s+(\w+)\s+AS\s+LIST\s+{(.*?)}\s*$", line)
    if matches:
        def_type = matches.group(1).lower()
        if def_type == "namespace":
            def_type = "namespaces"
        elif def_type == "annotation":
            def_type == "annotations"

        key = matches.group(2)
        val = matches.group(3)
        vals = convert_csv_str_to_list(val)

        if def_type in nanopubs_metadata:
            nanopubs_metadata[def_type][key] = vals
        else:
            nanopubs_metadata[def_type] = {key: vals}

    return nanopubs_metadata


def process_unset(line, annotations):
    """Process UNSET lines in BEL Script"""

    matches = re.match('UNSET\s+"?(.*?)"?\s*$', line)
    if matches:
        val = matches.group(1)
        if val == "ALL" or val == "STATEMENT_GROUP":
            annotations = {}
        elif re.match("{", val):
            vals = convert_csv_str_to_list(val)
            for val in vals:
                annotations.pop(val, None)
        else:
            annotations.pop(val, None)

    else:
        log.warn(f"Problem with UNSET line: {line}")

    return annotations


def process_set(line, annotations):
    """Convert annotations into nanopub_bel annotations format"""

    matches = re.match('SET\s+(\w+)\s*=\s*"?(.*?)"?\s*$', line)

    key = None
    if matches:
        key = matches.group(1)
        val = matches.group(2)

    if key == "STATEMENT_GROUP":
        annotations["statement_group"] = val
    elif key == "Citation":
        annotations["citation"] = process_citation(val)
    elif key.lower() == "support" or key.lower() == "evidence":
        annotations["evidence"] = val
    elif re.match("\s*{.*?}", val):
        vals = convert_csv_str_to_list(val)
        annotations[key] = vals
    else:
        annotations[key] = val

    return annotations


def set_single_line(lines):

    flag = False
    hold = ""

    for line in lines:
        if flag and re.match('.*"', line):
            line = hold + " " + line
            flag = False
            line = re.sub("\s+", " ", line)
            yield line
        elif flag:
            hold += " " + line.rstrip()

        elif re.match('SET\s*\w+\s*=\s*".*"', line):
            line = re.sub("\s+", " ", line)
            yield line
        elif re.match('SET\s*\w+\s*=\s*".*', line):
            hold = line.rstrip()
            flag = True
        else:
            line = re.sub("\s+", " ", line)
            yield line


def preprocess_belscript(lines):
    """ Convert any multi-line SET statements into single line SET statements"""

    set_flag = False
    for line in lines:
        if set_flag is False and re.match("SET", line):
            set_flag = True
            set_line = [line.rstrip()]
        # SET following SET
        elif set_flag and re.match("SET", line):
            yield f"{' '.join(set_line)}\n"
            set_line = [line.rstrip()]
        # Blank line following SET yields single line SET
        elif set_flag and re.match("\s+$", line):
            yield f"{' '.join(set_line)}\n"
            yield line
            set_flag = False

        # Append second, third, ... lines to SET
        elif set_flag:
            set_line.append(line.rstrip())
        else:
            yield line


def parse_belscript(lines):
    """Lines from the BELScript - can be an iterator or list

    yields Nanopubs in nanopubs_bel-1.0.0 format
    """

    nanopubs_metadata = {}
    annotations = {}
    assertions = []

    # # Turn a list into an iterator
    # if not isinstance(lines, collections.Iterator):
    #     lines = iter(lines)

    line_num = 0

    # for line in preprocess_belscript(lines):
    for line in set_single_line(lines):

        line_num += 1
        # Get rid of trailing comments
        line = re.sub("\/\/.*?$", "", line)

        line = line.rstrip()

        # Collapse continuation lines
        while re.search("\\\s*$", line):
            line = line.replace("\\", "") + next(lines)

        # Process lines #################################
        if re.match("\s*#", line) or re.match("\s*$", line):
            # Skip comments and empty lines
            continue
        elif re.match("SET DOCUMENT", line):
            nanopubs_metadata = process_documentline(line, nanopubs_metadata)
        elif re.match("DEFINE", line):
            nanopubs_metadata = process_definition(line, nanopubs_metadata)
        elif re.match("UNSET", line):

            # Process any assertions prior to changing annotations
            if assertions:
                yield yield_nanopub(assertions, annotations, line_num)
                assertions = []
            annotations = process_unset(line, annotations)

        elif re.match("SET", line):
            # Create nanopubs metadata prior to starting BEL Script statements section
            if nanopubs_metadata:
                yield yield_metadata(nanopubs_metadata)
                nanopubs_metadata = {}

            # Process any assertions prior to changing annotations
            if assertions:
                yield yield_nanopub(assertions, annotations, line_num)
                assertions = []

            annotations = process_set(line, annotations)

        else:
            assertions.append(line)

    # Catch any leftover bel statements
    yield_nanopub(assertions, annotations, line_num)


def main():

    with open("test.v2.bel", "r") as f:
        for doc in parse_belscript(f):
            print(json.dumps(doc, indent=4))

    quit()

    bel = 'proteinAbundance(HGNC:VHL) increases (proteinAbundance(HGNC:TNF) increases biologicalProcess(GOBP:"cell death"))'
    print(split_bel_stmt(bel))
    quit()

    citation_str = 'SET Citation = {"PubMed","Proc Natl Acad Sci U S A 1999 Feb 16 96(4) 1603-8","9990071","","",""}'
    print(process_citation(citation_str))
    quit()


if __name__ == "__main__":
    main()



import timy
import json
import gzip

from arango import ArangoError

from typing import IO

import bel.utils
import bel.db.arangodb as arangodb

from bel.Config import config

from structlog import get_logger

log = get_logger()


def load_orthologs(fo: IO, metadata: dict):
    """Load orthologs into ArangoDB

    Args:
        fo: file obj - orthologs file
        metadata: dict containing the metadata for orthologs
    """

    version = metadata["metadata"]["version"]

    # LOAD ORTHOLOGS INTO ArangoDB
    with timy.Timer("Load Orthologs") as timer:
        arango_client = arangodb.get_client()
        belns_db = arangodb.get_belns_handle(arango_client)
        arangodb.batch_load_docs(
            belns_db, orthologs_iterator(fo, version), on_duplicate="update"
        )

        log.info(
            "Load orthologs",
            elapsed=timer.elapsed,
            source=metadata["metadata"]["source"],
        )

        # Clean up old entries
        remove_old_ortholog_edges = f"""
            FOR edge in ortholog_edges
                FILTER edge.source == "{metadata["metadata"]["source"]}"
                FILTER edge.version != "{version}"
                REMOVE edge IN ortholog_edges
        """
        remove_old_ortholog_nodes = f"""
            FOR node in ortholog_nodes
                FILTER node.source == "{metadata["metadata"]["source"]}"
                FILTER node.version != "{version}"
                REMOVE node IN ortholog_nodes
        """
        arangodb.aql_query(belns_db, remove_old_ortholog_edges)
        arangodb.aql_query(belns_db, remove_old_ortholog_nodes)

    # Add metadata to resource metadata collection
    metadata["_key"] = f"Orthologs_{metadata['metadata']['source']}"
    try:
        belns_db.collection(arangodb.belns_metadata_name).insert(metadata)
    except ArangoError as ae:
        belns_db.collection(arangodb.belns_metadata_name).replace(metadata)


def orthologs_iterator(fo, version):
    """Ortholog node and edge iterator"""

    species_list = config["bel_resources"].get("species_list", [])

    fo.seek(0)
    with gzip.open(fo, "rt") as f:
        for line in f:
            edge = json.loads(line)
            if "metadata" in edge:
                source = edge["metadata"]["source"]
                continue

            if "ortholog" in edge:
                edge = edge["ortholog"]
                subj_tax_id = edge["subject"]["tax_id"]
                obj_tax_id = edge["object"]["tax_id"]

                # Skip if species not listed in species_list
                if species_list and subj_tax_id and subj_tax_id not in species_list:
                    continue
                if species_list and obj_tax_id and obj_tax_id not in species_list:
                    continue

                # Converted to ArangoDB legal chars for _key
                subj_key = arangodb.arango_id_to_key(edge["subject"]["id"])
                subj_id = edge["subject"]["id"]

                # Converted to ArangoDB legal chars for _key
                obj_key = arangodb.arango_id_to_key(edge["object"]["id"])
                obj_id = edge["object"]["id"]

                # Subject node
                yield (
                    arangodb.ortholog_nodes_name,
                    {
                        "_key": subj_key,
                        "name": subj_id,
                        "tax_id": edge["subject"]["tax_id"],
                        "source": source,
                        "version": version,
                    },
                )
                # Object node
                yield (
                    arangodb.ortholog_nodes_name,
                    {
                        "_key": obj_key,
                        "name": obj_id,
                        "tax_id": edge["object"]["tax_id"],
                        "source": source,
                        "version": version,
                    },
                )

                arango_edge = {
                    "_from": f"{arangodb.ortholog_nodes_name}/{subj_key}",
                    "_to": f"{arangodb.ortholog_nodes_name}/{obj_key}",
                    "_key": bel.utils._create_hash(f"{subj_id}>>{obj_id}"),
                    "type": "ortholog_to",
                    "source": source,
                    "version": version,
                }

                yield (arangodb.ortholog_edges_name, arango_edge)



from typing import Tuple
import re

import bel.db.elasticsearch
import bel.lang.belobj
from bel.Config import config

import structlog

log = structlog.getLogger(__name__)

es = bel.db.elasticsearch.get_client()


def convert_msg_to_html(msg):
    """Convert \n into a <BR> for an HTML formatted message"""

    msg = re.sub("\n", "<br />", msg, flags=re.MULTILINE)
    return msg


def validate(nanopub: dict, error_level: str = "WARNING") -> Tuple[str, str, str]:
    """Validate Nanopub

    Error Levels are similar to log levels - selecting WARNING includes both
    WARNING and ERROR, selecting ERROR just includes ERROR

    The validation result is a list of objects containing
        {
            'level': 'Warning|Error',
            'section': 'Assertion|Annotation|Structure',
            'label': '{Error|Warning}-{Assertion|Annotation|Structure}',  # to be used for faceting in Elasticsearch
            'index': idx,  # Index of Assertion or Annotation in Nanopub - starts at 0
            'msg': msg,  # Error or Warning message
        }

    Args:
        nanopub: nanopub record starting with nanopub...
        level: return WARNING or just ERROR?  defaults to warnings and errors
    Returns:
        list(tuples): [{'level': 'Warning', 'section': 'Assertion', 'label': 'Warning-Assertion', 'index': 0, 'msg': <msg>}]

    """

    # Validation results
    v = []

    bel_version = config["bel"]["lang"]["default_bel_version"]

    # Structural checks
    try:
        if not isinstance(nanopub["nanopub"]["assertions"], list):
            msg = "Assertions must be a list/array"
            v.append(
                {
                    "level": "Error",
                    "section": "Structure",
                    "label": "Error-Structure",
                    "msg": msg,
                    "msg_html": msg,
                }
            )
    except Exception as e:
        msg = 'Missing nanopub["nanopub"]["assertions"]'
        v.append(
            {
                "level": "Error",
                "section": "Structure",
                "label": "Error-Structure",
                "msg": msg,
                "msg_html": msg,
            }
        )

    try:
        if (
            "name" in nanopub["nanopub"]["type"]
            and "version" in nanopub["nanopub"]["type"]
        ):
            pass
        if nanopub["nanopub"]["type"]["name"].upper() == "BEL":
            bel_version = nanopub["nanopub"]["type"]["version"]

    except Exception as e:
        msg = 'Missing or badly formed type - must have nanopub["nanopub"]["type"] = {"name": <name>, "version": <version}'
        v.append(
            {
                "level": "Error",
                "section": "Structure",
                "label": "Error-Structure",
                "msg": msg,
                "msg_html": msg,
            }
        )

    try:
        for key in ["uri", "database", "reference"]:
            if key in nanopub["nanopub"]["citation"]:
                break
        else:
            msg = 'nanopub["nanopub"]["citation"] must have either a uri, database or reference key.'
            v.append(
                {
                    "level": "Error",
                    "section": "Structure",
                    "label": "Error-Structure",
                    "msg": msg,
                    "msg_html": msg,
                }
            )
    except Exception as e:
        msg = 'nanopub["nanopub"] must have a "citation" key with either a uri, database or reference key.'
        v.append(
            {
                "level": "Error",
                "section": "Structure",
                "label": "Error-Structure",
                "msg": msg,
                "msg_html": msg,
            }
        )

    # Assertion checks
    if "assertions" in nanopub["nanopub"]:
        for idx, assertion in enumerate(nanopub["nanopub"]["assertions"]):
            bo = bel.lang.belobj.BEL(
                bel_version, config["bel_api"]["servers"]["api_url"]
            )
            belstr = f'{assertion.get("subject")} {assertion.get("relation", "")} {assertion.get("object", "")}'
            belstr = belstr.replace("None", "")
            try:
                messages = (
                    bo.parse(belstr)
                    .semantic_validation(error_level=error_level)
                    .validation_messages
                )

                for message in messages:
                    (level, msg) = message
                    if error_level == "ERROR":
                        if level == "ERROR":
                            v.append(
                                {
                                    "level": f"{level.title()}",
                                    "section": "Assertion",
                                    "label": f"{level.title()}-Assertion",
                                    "index": idx,
                                    "msg": msg,
                                    "msg_html": convert_msg_to_html(msg),
                                }
                            )
                    else:
                        v.append(
                            {
                                "level": f"{level.title()}",
                                "section": "Assertion",
                                "label": f"{level.title()}-Assertion",
                                "index": idx,
                                "msg": msg,
                                "msg_html": convert_msg_to_html(msg),
                            }
                        )

            except Exception as e:
                msg = f"Could not parse: {belstr}"
                v.append(
                    {
                        "level": "Error",
                        "section": "Assertion",
                        "label": "Error-Assertion",
                        "index": idx,
                        "msg": msg,
                        "msg_html": msg,
                    }
                )
                log.exception(f"Could not parse: {belstr}")

    # Annotation checks
    if error_level == "WARNING":
        for idx, annotation in enumerate(nanopub["nanopub"].get("annotations", [])):
            term_type = annotation["type"]
            term_id = annotation["id"]
            # term_label = annotation['label']
            log.info(f"Annotation: {term_type}  ID: {term_id}")

            search_body = {
                "_source": ["src_id", "id", "name", "label", "annotation_types"],
                "query": {"term": {"id": term_id}},
            }

            results = es.search(index="terms", doc_type="term", body=search_body)
            if len(results["hits"]["hits"]) > 0:
                result = results["hits"]["hits"][0]["_source"]
                if term_type not in result["annotation_types"]:
                    msg = f'Annotation type: {term_type} for {term_id} does not match annotation types in database: {result["annotation_types"]}'
                    v.append(
                        {
                            "level": "Warning",
                            "section": "Annotation",
                            "index": idx,
                            "label": "Warning-Annotation",
                            "msg": msg,
                            "msg_html": msg,
                        }
                    )
            else:
                msg = f"Annotation term: {term_id} not found in database"
                v.append(
                    {
                        "level": "Warning",
                        "section": "Annotation",
                        "index": idx,
                        "label": "Warning-Annotation",
                        "msg": msg,
                        "msg_html": msg,
                    }
                )

    return v



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""


"""

import json
import yaml
import re
import copy
import sys
import click
from typing import Mapping, Any, List, Iterable, Tuple
import gzip

import logging

log = logging.getLogger(__name__)


def read_nanopubs(fn: str) -> Iterable[Mapping[str, Any]]:
    """Read file and generate nanopubs

    If filename has *.gz, will read as a gzip file
    If filename has *.jsonl*, will parsed as a JSONLines file
    IF filename has *.json*, will be parsed as a JSON file
    If filename has *.yaml* or *.yml*,  will be parsed as a YAML file

    Args:
        filename (str): filename to read nanopubs from

    Returns:
        Generator[Mapping[str, Any]]: generator of nanopubs in nanopub_bel JSON Schema format
    """

    jsonl_flag, json_flag, yaml_flag = False, False, False
    if fn == "-" or "jsonl" in fn:
        jsonl_flag = True
    elif "json" in fn:
        json_flag = True
    elif re.search("ya?ml", fn):
        yaml_flag = True
    else:
        log.error("Do not recognize nanopub file format - neither json nor jsonl format.")
        return {}

    try:
        if re.search("gz$", fn):
            f = gzip.open(fn, "rt")
        else:
            try:
                f = click.open_file(fn, mode="rt")
            except Exception as e:
                log.info(f"Can not open file {fn}  Error: {e}")
                quit()

        if jsonl_flag:
            for line in f:
                yield json.loads(line)
        elif json_flag:
            nanopubs = json.load(f)
            for nanopub in nanopubs:
                yield nanopub
        elif yaml_flag:
            nanopubs = yaml.load(f, Loader=yaml.SafeLoader)
            for nanopub in nanopubs:
                yield nanopub

    except Exception as e:
        log.error(f"Could not open file: {fn}")


def create_nanopubs_fh(output_fn: str):
    """Create Nanopubs output filehandle

    \b
    If output fn is '-' will write JSONlines to STDOUT
    If output fn has *.gz, will written as a gzip file
    If output fn has *.jsonl*, will written as a JSONLines file
    IF output fn has *.json*, will be written as a JSON file
    If output fn has *.yaml* or *.yml*,  will be written as a YAML file

    Args:
        output_fn: Name of output file

    Returns:
        (filehandle, yaml_flag, jsonl_flag, json_flag)
    """

    # output file
    # set output flags
    json_flag, jsonl_flag, yaml_flag = False, False, False
    if output_fn:
        if re.search("gz$", output_fn):
            out_fh = gzip.open(output_fn, "wt")
        else:
            out_fh = click.open_file(output_fn, mode="wt")

        if re.search("ya?ml", output_fn):
            yaml_flag = True
        elif "jsonl" in output_fn or "-" == output_fn:
            jsonl_flag = True
        elif "json" in output_fn:
            json_flag = True

    else:
        out_fh = sys.stdout

    return (out_fh, yaml_flag, jsonl_flag, json_flag)


def read_edges(fn):

    if not fn:
        return []
    jsonl_flag, json_flag, yaml_flag = False, False, False
    if "jsonl" in fn:
        jsonl_flag = True
    elif "json" in fn:
        json_flag = True
    elif re.search("ya?ml", fn):
        yaml_flag = True
    else:
        log.error("Do not recognize edge file format - neither json nor jsonl format.")
        return []

    try:
        if re.search("gz$", fn):
            f = gzip.open(fn, "rt")
        else:
            f = open(fn, "rt")

        if jsonl_flag:
            for line in f:
                edges = json.loads(line)
                for edge in edges:
                    yield edge
        elif json_flag:
            edges = json.load(f)
            for edge in edges:
                yield edge
        elif yaml_flag:
            edges = yaml.load_all(f, Loader=yaml.SafeLoader)
            for edge in edges:
                yield edge

    except Exception as e:
        log.error(f"Could not open file: {fn}")


def write_edges(
    edges: Mapping[str, Any],
    filename: str,
    jsonlines: bool = False,
    gzipflag: bool = False,
    yaml: bool = False,
):
    """Write edges to file

    Args:
        edges (Mapping[str, Any]): in edges JSON Schema format
        filename (str): filename to write
        jsonlines (bool): output in JSONLines format?
        gzipflag (bool): create gzipped file?
        yaml (bool): create yaml file?
    """
    pass


def edges_to_jgf(fn, edges):

    jgf_nodes = []
    jgf_edges = []
    for edge in edges:
        if "edge" in edge:
            jgf_nodes.append({"id": edge["edge"]["subject"]["name"]})
            jgf_nodes.append({"id": edge["edge"]["object"]["name"]})
            jgf_edges.append(
                {
                    "source": edge["edge"]["subject"]["name"],
                    "target": edge["edge"]["object"]["name"],
                    "relation": edge["edge"]["relation"]["name"],
                }
            )

    with open(fn, "wt") as f:
        graph = {
            "graph": {
                "label": "BEL Pipeline Edges",
                "type": "BEL Edges",
                "nodes": copy.deepcopy(jgf_nodes),
                "edges": copy.deepcopy(jgf_edges),
            }
        }

        f.write("{}\n".format(json.dumps(graph, indent=4)))


def main():
    pass


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = "bel"
DESCRIPTION = "BEL related functionality"
URL = "https://github.com/belbio/bel"
EMAIL = "whayes@biodati.com"
AUTHOR = "William Hayes"

# What packages are required for this module to be executed?
REQUIRED = [
    "cityhash",
    "click",
    "elasticsearch>=6,<7",
    "fastcache",
    "Jinja2",
    "jsonschema",
    "lxml",
    "python-arango",
    "python-dateutil",
    "python-json-logger",
    "pyyaml>5",
    "requests",
    "requests-cache",
    "structlog",
    "timy",
    "TatSu",
    "ulid-py",
]

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.rst' is present in your MANIFEST.in file!
with io.open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = "\n" + f.read()

# Load the package's __version__.py module as a dictionary.
about = {}
with open(os.path.join(here, NAME, "__version__.py")) as f:
    exec(f.read(), about)


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPi via Twine…")
        os.system("twine upload dist/*")

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    packages=find_packages(include=["bel"], exclude=["tests"]),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],
    entry_points={
        "console_scripts": [
            "belc=bel.scripts:belc",
            "belspec_yaml2json=bel.lang.bel_specification:belspec_yaml2json",
        ]
    },
    install_requires=REQUIRED,
    python_requires="~=3.6",
    include_package_data=True,
    license="Apache2",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Intended Audience :: Science/Research",
        "Development Status :: 4 - Beta",
    ],
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand},
)
